# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access, too-many-lines

import datetime
from typing import Any, Dict, List, Optional, Union

import isodate
from typing_extensions import Literal

from azure.ai.ml._exception_helper import log_and_raise_error
from azure.ai.ml._restclient.v2023_06_01_preview.models import AllFeatures as RestAllFeatures
from azure.ai.ml._restclient.v2023_06_01_preview.models import CustomMonitoringSignal as RestCustomMonitoringSignal
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    DataDriftMonitoringSignal as RestMonitoringDataDriftSignal,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    DataQualityMonitoringSignal as RestMonitoringDataQualitySignal,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    FeatureAttributionDriftMonitoringSignal as RestFeatureAttributionDriftMonitoringSignal,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import FeatureSubset as RestFeatureSubset
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    GenerationSafetyQualityMonitoringSignal as RestGenerationSafetyQualityMonitoringSignal,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    GenerationTokenStatisticsSignal as RestGenerationTokenStatisticsSignal,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import ModelPerformanceSignal as RestModelPerformanceSignal
from azure.ai.ml._restclient.v2023_06_01_preview.models import MonitoringDataSegment as RestMonitoringDataSegment
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    MonitoringFeatureFilterBase as RestMonitoringFeatureFilterBase,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import MonitoringInputDataBase as RestMonitoringInputData
from azure.ai.ml._restclient.v2023_06_01_preview.models import MonitoringNotificationMode
from azure.ai.ml._restclient.v2023_06_01_preview.models import MonitoringSignalBase as RestMonitoringSignalBase
from azure.ai.ml._restclient.v2023_06_01_preview.models import MonitoringSignalType
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    MonitoringWorkspaceConnection as RestMonitoringWorkspaceConnection,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    PredictionDriftMonitoringSignal as RestPredictionDriftMonitoringSignal,
)
from azure.ai.ml._restclient.v2023_06_01_preview.models import (
    TopNFeaturesByAttribution as RestTopNFeaturesByAttribution,
)
from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml.constants._monitoring import (
    ALL_FEATURES,
    MonitorDatasetContext,
    MonitorFeatureDataType,
    MonitorSignalType,
)
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job._input_output_helpers import (
    from_rest_inputs_to_dataset_literal,
    to_rest_dataset_literal_inputs,
)
from azure.ai.ml.entities._mixins import RestTranslatableMixin
from azure.ai.ml.entities._monitoring.input_data import FixedInputData, StaticInputData, TrailingInputData
from azure.ai.ml.entities._monitoring.thresholds import (
    CustomMonitoringMetricThreshold,
    DataDriftMetricThreshold,
    DataQualityMetricThreshold,
    FeatureAttributionDriftMetricThreshold,
    GenerationSafetyQualityMonitoringMetricThreshold,
    GenerationTokenStatisticsMonitorMetricThreshold,
    MetricThreshold,
    ModelPerformanceMetricThreshold,
    PredictionDriftMetricThreshold,
)
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationErrorType, ValidationException


class DataSegment(RestTranslatableMixin):
    """Data segment for monitoring.

    :keyword feature_name: The feature to segment the data on.
    :paramtype feature_name: str
    :keyword feature_values: A list of values for the given segmented feature to filter.
    :paramtype feature_values: List[str]
    """

    def __init__(
        self,
        *,
        feature_name: Optional[str] = None,
        feature_values: Optional[List[str]] = None,
    ) -> None:
        self.feature_name = feature_name
        self.feature_values = feature_values

    def _to_rest_object(self) -> RestMonitoringDataSegment:
        return RestMonitoringDataSegment(feature=self.feature_name, values=self.feature_values)

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringDataSegment) -> "DataSegment":
        return cls(
            feature_name=obj.feature,
            feature_values=obj.values,
        )


class MonitorFeatureFilter(RestTranslatableMixin):
    """Monitor feature filter

    :keyword top_n_feature_importance: The number of top features to include. Defaults to 10.
    :paramtype top_n_feature_importance: int
    """

    def __init__(
        self,
        *,
        top_n_feature_importance: int = 10,
    ) -> None:
        self.top_n_feature_importance = top_n_feature_importance

    def _to_rest_object(self) -> RestTopNFeaturesByAttribution:
        return RestTopNFeaturesByAttribution(
            top=self.top_n_feature_importance,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestTopNFeaturesByAttribution) -> "MonitorFeatureFilter":
        return cls(top_n_feature_importance=obj.top)


class BaselineDataRange:
    """Baseline data range for monitoring.

    This class is used when initializing a data_window for a ReferenceData object.
    For trailing input, set lookback_window_size and lookback_window_offset to a desired value.
    For static input, set window_start and window_end to a desired value.
    """

    def __init__(
        self,
        *,
        window_start: Optional[str] = None,
        window_end: Optional[str] = None,
        lookback_window_size: Optional[str] = None,
        lookback_window_offset: Optional[str] = None,
    ):
        self.window_start = window_start
        self.window_end = window_end
        self.lookback_window_size = lookback_window_size
        self.lookback_window_offset = lookback_window_offset


class ProductionData(RestTranslatableMixin):
    """Production Data

    :param input_data: The data for which drift will be calculated
    :type Input: ~azure.ai.ml.entities._input_outputs
    :param data_context: The context of the input dataset. Possible values
        include: model_inputs, model_outputs, training, test, validation, ground_truth
    :type MonitorDatasetContext: ~azure.ai.ml.constants.MonitorDatasetContext
    :param pre_processing_component: ARM resource ID of the component resource used to
        preprocess the data.
    :type pre_processing_component: string
    :param data_window: The number of days or a time frame that a singal monitor looks back over the target.
    :type data_window_size: BaselineDataRange
    """

    def __init__(
        self,
        *,
        input_data: Input,
        data_context: Optional[MonitorDatasetContext] = None,
        pre_processing_component: Optional[str] = None,
        data_window: Optional[BaselineDataRange] = None,
        data_column_names: Optional[Dict[str, str]] = None,
    ):
        self.input_data = input_data
        self.data_context = data_context
        self.pre_processing_component = pre_processing_component
        self.data_window = data_window
        self.data_column_names = data_column_names

    def _to_rest_object(self, **kwargs: Any) -> RestMonitoringInputData:
        self._validate()
        default_data_window_size = kwargs.get("default_data_window_size")
        if self.data_window is None:
            self.data_window = BaselineDataRange(
                lookback_window_size=default_data_window_size, lookback_window_offset="P0D"
            )
        if self.data_window.lookback_window_size in ["default", None]:
            self.data_window.lookback_window_size = default_data_window_size
        uri = self.input_data.path
        job_type = self.input_data.type
        monitoring_input_data = TrailingInputData(
            data_context=self.data_context,
            target_columns=self.data_column_names,
            job_type=job_type,
            uri=uri,
            pre_processing_component_id=self.pre_processing_component,
            window_size=self.data_window.lookback_window_size,
            window_offset=(
                self.data_window.lookback_window_offset
                if self.data_window.lookback_window_offset is not None
                else "P0D"
            ),
        )
        return monitoring_input_data._to_rest_object()

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringInputData) -> "ProductionData":
        data_window = BaselineDataRange(
            lookback_window_size=isodate.duration_isoformat(obj.window_size),
            lookback_window_offset=isodate.duration_isoformat(obj.window_offset),
        )
        return cls(
            input_data=Input(
                path=obj.uri,
                type=obj.job_input_type,
            ),
            data_context=obj.data_context,
            pre_processing_component=obj.preprocessing_component_id,
            data_window=data_window,
            data_column_names=obj.columns,
        )

    def _validate(self) -> None:
        if self.data_window:
            if self.data_window.window_start or self.data_window.window_end:
                msg = "ProductionData only accepts lookback_window_size and lookback_window_offset."
                err = ValidationException(
                    message=msg,
                    target=ErrorTarget.MODEL_MONITORING,
                    no_personal_data_message=msg,
                    error_category=ErrorCategory.USER_ERROR,
                    error_type=ValidationErrorType.MISSING_FIELD,
                )
                log_and_raise_error(err)


class ReferenceData(RestTranslatableMixin):
    """Reference Data

    :param input_data: The data for which drift will be calculated
    :type Input: ~azure.ai.ml.entities._input_outputs
    :param data_context: The context of the input dataset. Possible values
        include: model_inputs, model_outputs, training, test, validation, ground_truth
    :type MonitorDatasetContext: ~azure.ai.ml.constants.MonitorDatasetContext
    :param pre_processing_component: ARM resource ID of the component resource used to
        preprocess the data.
    :type pre_processing_component: string
    :param target_column_name: The name of the target column in the dataset.
    :type target_column_name: string
    :param data_window: The number of days or a time frame that a single monitor looks back over the target.
    :type data_window_size: BaselineDataRange
    """

    def __init__(
        self,
        *,
        input_data: Input,
        data_context: Optional[MonitorDatasetContext] = None,
        pre_processing_component: Optional[str] = None,
        data_window: Optional[BaselineDataRange] = None,
        data_column_names: Optional[Dict[str, str]] = None,
    ):
        self.input_data = input_data
        self.data_context = data_context
        self.pre_processing_component = pre_processing_component
        self.data_window = data_window
        self.data_column_names = data_column_names

    def _to_rest_object(self, **kwargs: Any) -> RestMonitoringInputData:
        default_data_window = kwargs.get("default_data_window")
        ref_data_window_size = kwargs.get("ref_data_window_size")
        if self.data_window is not None:
            if self.data_window.lookback_window_size is not None:
                if self.data_window.lookback_window_size == "default":
                    self.data_window.lookback_window_size = ref_data_window_size
                if self.data_window.lookback_window_offset == "default":
                    self.data_window.lookback_window_offset = default_data_window
                return TrailingInputData(
                    data_context=self.data_context,
                    target_columns=self.data_column_names,
                    job_type=self.input_data.type,
                    uri=self.input_data.path,
                    pre_processing_component_id=self.pre_processing_component,
                    window_size=self.data_window.lookback_window_size,
                    window_offset=(
                        self.data_window.lookback_window_offset
                        if self.data_window.lookback_window_offset is not None
                        else "P0D"
                    ),
                )._to_rest_object()
            if self.data_window.window_start is not None and self.data_window.window_end is not None:
                return StaticInputData(
                    data_context=self.data_context,
                    target_columns=self.data_column_names,
                    job_type=self.input_data.type,
                    uri=self.input_data.path,
                    pre_processing_component_id=self.pre_processing_component,
                    window_start=self.data_window.window_start,
                    window_end=self.data_window.window_end,
                )._to_rest_object()

        return FixedInputData(
            data_context=self.data_context,
            target_columns=self.data_column_names,
            job_type=self.input_data.type,
            uri=self.input_data.path,
        )._to_rest_object()

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringInputData) -> "ReferenceData":
        data_window = None
        if obj.input_data_type == "Static":
            data_window = BaselineDataRange(
                window_start=datetime.datetime.strftime(obj.window_start, "%Y-%m-%d"),
                window_end=datetime.datetime.strftime(obj.window_end, "%Y-%m-%d"),
            )
        if obj.input_data_type == "Trailing":
            data_window = BaselineDataRange(
                lookback_window_size=isodate.duration_isoformat(obj.window_size),
                lookback_window_offset=isodate.duration_isoformat(obj.window_offset),
            )

        return cls(
            input_data=Input(
                path=obj.uri,
                type=obj.job_input_type,
            ),
            data_context=obj.data_context,
            pre_processing_component=obj.preprocessing_component_id if obj.input_data_type != "Fixed" else None,
            data_window=data_window,
            data_column_names=obj.columns,
        )


class MonitoringSignal(RestTranslatableMixin):
    """
    Base class for monitoring signals.

    This class should not be instantiated directly. Instead, use one of its subclasses.

    :keyword baseline_dataset: The baseline dataset definition for monitor input.
    :paramtype baseline_dataset: ~azure.ai.ml.entities.MonitorInputData
    :keyword metric_thresholds: The metric thresholds for the signal.
    :paramtype metric_thresholds: Union[
        ~azure.ai.ml.entities.DataDriftMetricThreshold,
        ~azure.ai.ml.entities.DataQualityMetricThreshold,
        ~azure.ai.ml.entities.PredictionDriftMetricThreshold,
        ~azure.ai.ml.entities.FeatureAttributionDriftMetricThreshold,
        ~azure.ai.ml.entities.CustomMonitoringMetricThreshold,
        ~azure.ai.ml.entities.GenerationSafetyQualityMonitoringMetricThreshold,
        List[Union[
            ~azure.ai.ml.entities.DataDriftMetricThreshold,
            ~azure.ai.ml.entities.DataQualityMetricThreshold,
            ~azure.ai.ml.entities.PredictionDriftMetricThreshold,
            ~azure.ai.ml.entities.FeatureAttributionDriftMetricThreshold,
            ~azure.ai.ml.entities.CustomMonitoringMetricThreshold,
            ~azure.ai.ml.entities.GenerationSafetyQualityMonitoringMetricThreshold,

        ]]]
    :keyword alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    """

    def __init__(
        self,
        *,
        production_data: Optional[ProductionData] = None,
        reference_data: Optional[ReferenceData] = None,
        metric_thresholds: Optional[Union[MetricThreshold, List[MetricThreshold]]],
        properties: Optional[Dict[str, str]] = None,
        alert_enabled: bool = False,
    ):
        self.production_data = production_data
        self.reference_data = reference_data
        self.metric_thresholds = metric_thresholds
        self.alert_enabled = alert_enabled
        self.properties = properties

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringSignalBase) -> Optional[  # pylint: disable=too-many-return-statements
        Union[
            "DataDriftSignal",
            "DataQualitySignal",
            "PredictionDriftSignal",
            "ModelPerformanceSignal",
            "FeatureAttributionDriftSignal",
            "CustomMonitoringSignal",
            "GenerationSafetyQualitySignal",
            "GenerationTokenStatisticsSignal",
        ]
    ]:
        if obj.signal_type == MonitoringSignalType.DATA_DRIFT:
            return DataDriftSignal._from_rest_object(obj)
        if obj.signal_type == MonitoringSignalType.DATA_QUALITY:
            return DataQualitySignal._from_rest_object(obj)
        if obj.signal_type == MonitoringSignalType.PREDICTION_DRIFT:
            return PredictionDriftSignal._from_rest_object(obj)
        if obj.signal_type == "ModelPerformanceSignalBase":
            return ModelPerformanceSignal._from_rest_object(obj)
        if obj.signal_type == MonitoringSignalType.FEATURE_ATTRIBUTION_DRIFT:
            return FeatureAttributionDriftSignal._from_rest_object(obj)
        if obj.signal_type == MonitoringSignalType.CUSTOM:
            return CustomMonitoringSignal._from_rest_object(obj)
        if obj.signal_type == MonitoringSignalType.GENERATION_SAFETY_QUALITY:
            return GenerationSafetyQualitySignal._from_rest_object(obj)
        if obj.signal_type == MonitoringSignalType.MODEL_PERFORMANCE:
            return ModelPerformanceSignal._from_rest_object(obj)
        if obj.signal_type == MonitoringSignalType.GENERATION_TOKEN_STATISTICS:
            return GenerationTokenStatisticsSignal._from_rest_object(obj)

        return None


class DataSignal(MonitoringSignal):
    """Base class for data signals.

    This class should not be instantiated directly. Instead, use one of its subclasses.

    :keyword baseline_dataset: The baseline dataset definition for monitor input.
    :paramtype baseline_dataset: ~azure.ai.ml.entities.MonitorInputData
    :keyword features: The features to include in the signal.
    :paramtype features: Union[List[str], ~azure.ai.ml.entities.MonitorFeatureFilter, Literal[ALL_FEATURES]]
    :keyword metric_thresholds: The metric thresholds for the signal.
    :paramtype metric_thresholds: List[Union[
        ~azure.ai.ml.entities.DataDriftMetricThreshold,
        ~azure.ai.ml.entities.DataQualityMetricThreshold,
        ~azure.ai.ml.entities.PredictionDriftMetricThreshold,
        ~azure.ai.ml.entities.FeatureAttributionDriftMetricThreshold,
        ~azure.ai.ml.entities.CustomMonitoringMetricThreshold,
        ~azure.ai.ml.entities.GenerationSafetyQualityMonitoringMetricThreshold,

    ]]
    :keyword alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    """

    def __init__(
        self,
        *,
        production_data: Optional[ProductionData] = None,
        reference_data: Optional[ReferenceData] = None,
        features: Optional[Union[List[str], MonitorFeatureFilter, Literal["all_features"]]] = None,
        feature_type_override: Optional[Dict[str, Union[str, MonitorFeatureDataType]]] = None,
        metric_thresholds: Optional[Union[MetricThreshold, List[MetricThreshold]]],
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            production_data=production_data,
            reference_data=reference_data,
            metric_thresholds=metric_thresholds,
            alert_enabled=alert_enabled,
            properties=properties,
        )
        self.features = features
        self.feature_type_override = feature_type_override


class DataDriftSignal(DataSignal):
    """Data drift signal.

    :ivar type: The type of the signal, set to "data_drift" for this class.
    :vartype type: str
    :param production_data: The data for which drift will be calculated
    :paramtype production_data: ~azure.ai.ml.entities.ProductionData
    :param reference_data: The data to calculate drift against
    :paramtype reference_data: ~azure.ai.ml.entities.ReferenceData
    :param metric_thresholds: Metrics to calculate and their associated thresholds
    :paramtype metric_thresholds: ~azure.ai.ml.entities.DataDriftMetricThreshold
    :param alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    :param data_segment: The data segment used for scoping on a subset of the data population.
    :paramtype data_segment: ~azure.ai.ml.entities.DataSegment
    :keyword features: The feature filter identifying which feature(s) to calculate drift over.
    :paramtype features: Union[List[str], ~azure.ai.ml.entities.MonitorFeatureFilter, Literal['all_features']]
    :param feature_type_override: Dictionary of features and what they should be overridden to.
    :paramtype feature_type_override: dict[str, str]
    :param properties: Dictionary of additional properties.
    :paramtype properties: dict[str, str]
    """

    def __init__(
        self,
        *,
        production_data: Optional[ProductionData] = None,
        reference_data: Optional[ReferenceData] = None,
        features: Optional[Union[List[str], MonitorFeatureFilter, Literal["all_features"]]] = None,
        feature_type_override: Optional[Dict[str, Union[str, MonitorFeatureDataType]]] = None,
        metric_thresholds: Optional[Union[DataDriftMetricThreshold, List[MetricThreshold]]] = None,
        alert_enabled: bool = False,
        data_segment: Optional[DataSegment] = None,
        properties: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            production_data=production_data,
            reference_data=reference_data,
            metric_thresholds=metric_thresholds,
            features=features,
            feature_type_override=feature_type_override,
            alert_enabled=alert_enabled,
            properties=properties,
        )
        self.type = MonitorSignalType.DATA_DRIFT
        self.data_segment = data_segment

    def _to_rest_object(self, **kwargs: Any) -> RestMonitoringDataDriftSignal:
        default_data_window_size = kwargs.get("default_data_window_size")
        ref_data_window_size = kwargs.get("ref_data_window_size")
        if self.production_data is not None and self.production_data.data_window is None:
            self.production_data.data_window = BaselineDataRange(lookback_window_size=default_data_window_size)
        rest_features = _to_rest_features(self.features) if self.features else None
        return RestMonitoringDataDriftSignal(
            production_data=(
                self.production_data._to_rest_object(default_data_window_size=default_data_window_size)
                if self.production_data is not None
                else None
            ),
            reference_data=(
                self.reference_data._to_rest_object(
                    default_data_window=default_data_window_size, ref_data_window_size=ref_data_window_size
                )
                if self.reference_data is not None
                else None
            ),
            features=rest_features,
            feature_data_type_override=self.feature_type_override,
            metric_thresholds=(
                self.metric_thresholds._to_rest_object()
                if isinstance(self.metric_thresholds, MetricThreshold)
                else None
            ),
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            data_segment=self.data_segment._to_rest_object() if self.data_segment else None,
            properties=self.properties,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringDataDriftSignal) -> "DataDriftSignal":
        return cls(
            production_data=ProductionData._from_rest_object(obj.production_data),
            reference_data=ReferenceData._from_rest_object(obj.reference_data),
            features=_from_rest_features(obj.features),
            feature_type_override=obj.feature_data_type_override,
            metric_thresholds=DataDriftMetricThreshold._from_rest_object(obj.metric_thresholds),
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
            data_segment=DataSegment._from_rest_object(obj.data_segment) if obj.data_segment else None,
            properties=obj.properties,
        )

    @classmethod
    def _get_default_data_drift_signal(cls) -> "DataDriftSignal":
        return cls(
            features=ALL_FEATURES,  # type: ignore[arg-type]
            metric_thresholds=DataDriftMetricThreshold._get_default_thresholds(),
        )


class PredictionDriftSignal(MonitoringSignal):
    """Prediction drift signal.

    :ivar type: The type of the signal, set to "prediction_drift" for this class.
    :vartype type: str
    :param production_data: The data for which drift will be calculated
    :paramtype production_data: ~azure.ai.ml.entities.ProductionData
    :param reference_data: The data to calculate drift against
    :paramtype reference_data: ~azure.ai.ml.entities.ReferenceData
    :param metric_thresholds: Metrics to calculate and their associated thresholds
    :paramtype metric_thresholds: ~azure.ai.ml.entities.DataDriftMetricThreshold
    :param alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    :param properties: Dictionary of additional properties.
    :paramtype properties: dict[str, str]
    """

    def __init__(
        self,
        *,
        production_data: Optional[ProductionData] = None,
        reference_data: Optional[ReferenceData] = None,
        metric_thresholds: PredictionDriftMetricThreshold,
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            production_data=production_data,
            reference_data=reference_data,
            metric_thresholds=metric_thresholds,
            alert_enabled=alert_enabled,
            properties=properties,
        )
        self.type = MonitorSignalType.PREDICTION_DRIFT

    def _to_rest_object(self, **kwargs: Any) -> RestPredictionDriftMonitoringSignal:
        default_data_window_size = kwargs.get("default_data_window_size")
        ref_data_window_size = kwargs.get("ref_data_window_size")
        if self.production_data is not None and self.production_data.data_window is None:
            self.production_data.data_window = BaselineDataRange(lookback_window_size=default_data_window_size)
        return RestPredictionDriftMonitoringSignal(
            production_data=(
                self.production_data._to_rest_object(default_data_window_size=default_data_window_size)
                if self.production_data is not None
                else None
            ),
            reference_data=(
                self.reference_data._to_rest_object(
                    default_data_window=default_data_window_size, ref_data_window_size=ref_data_window_size
                )
                if self.reference_data is not None
                else None
            ),
            metric_thresholds=(
                self.metric_thresholds._to_rest_object()
                if isinstance(self.metric_thresholds, MetricThreshold)
                else None
            ),
            properties=self.properties,
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            model_type="classification",
        )

    @classmethod
    def _from_rest_object(cls, obj: RestPredictionDriftMonitoringSignal) -> "PredictionDriftSignal":
        return cls(
            production_data=ProductionData._from_rest_object(obj.production_data),
            reference_data=ReferenceData._from_rest_object(obj.reference_data),
            metric_thresholds=PredictionDriftMetricThreshold._from_rest_object(obj.metric_thresholds),
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
            properties=obj.properties,
        )

    @classmethod
    def _get_default_prediction_drift_signal(cls) -> "PredictionDriftSignal":
        return cls(
            metric_thresholds=PredictionDriftMetricThreshold._get_default_thresholds(),
        )


class DataQualitySignal(DataSignal):
    """Data quality signal

    :ivar type: The type of the signal. Set to "data_quality" for this class.
    :vartype type: str
    :param production_data: The data for which drift will be calculated
    :paramtype production_data: ~azure.ai.ml.entities.ProductionData
    :param reference_data: The data to calculate drift against
    :paramtype reference_data: ~azure.ai.ml.entities.ReferenceData
    :param metric_thresholds: Metrics to calculate and their associated thresholds
    :paramtype metric_thresholds: ~azure.ai.ml.entities.DataDriftMetricThreshold
    :param alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    :keyword features: The feature filter identifying which feature(s) to calculate drift over.
    :paramtype features: Union[List[str], ~azure.ai.ml.entities.MonitorFeatureFilter, Literal['all_features']]
    :param feature_type_override: Dictionary of features and what they should be overridden to.
    :paramtype feature_type_override: dict[str, str]
    :param properties: Dictionary of additional properties.
    :paramtype properties: dict[str, str]
    """

    def __init__(
        self,
        *,
        production_data: Optional[ProductionData] = None,
        reference_data: Optional[ReferenceData] = None,
        features: Optional[Union[List[str], MonitorFeatureFilter, Literal["all_features"]]] = None,
        feature_type_override: Optional[Dict[str, Union[str, MonitorFeatureDataType]]] = None,
        metric_thresholds: Optional[Union[MetricThreshold, List[MetricThreshold]]] = None,
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            production_data=production_data,
            reference_data=reference_data,
            metric_thresholds=metric_thresholds,
            features=features,
            feature_type_override=feature_type_override,
            alert_enabled=alert_enabled,
            properties=properties,
        )
        self.type = MonitorSignalType.DATA_QUALITY

    def _to_rest_object(self, **kwargs: Any) -> RestMonitoringDataQualitySignal:
        default_data_window_size = kwargs.get("default_data_window_size")
        ref_data_window_size = kwargs.get("ref_data_window_size")
        if self.production_data is not None and self.production_data.data_window is None:
            self.production_data.data_window = BaselineDataRange(
                lookback_window_size=default_data_window_size,
            )
        rest_features = _to_rest_features(self.features) if self.features else None
        rest_metrics = (
            # TODO: Bug Item number: 2883365
            _to_rest_data_quality_metrics(
                self.metric_thresholds.numerical, self.metric_thresholds.categorical  # type: ignore
            )
            if isinstance(self.metric_thresholds, MetricThreshold)
            else None
        )
        return RestMonitoringDataQualitySignal(
            production_data=(
                self.production_data._to_rest_object(default_data_window_size=default_data_window_size)
                if self.production_data is not None
                else None
            ),
            reference_data=(
                self.reference_data._to_rest_object(
                    default_data_window=default_data_window_size, ref_data_window_size=ref_data_window_size
                )
                if self.reference_data is not None
                else None
            ),
            features=rest_features,
            feature_data_type_override=self.feature_type_override,
            metric_thresholds=rest_metrics,
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            properties=self.properties,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringDataQualitySignal) -> "DataQualitySignal":
        return cls(
            production_data=ProductionData._from_rest_object(obj.production_data),
            reference_data=ReferenceData._from_rest_object(obj.reference_data),
            features=_from_rest_features(obj.features),
            feature_type_override=obj.feature_data_type_override,
            metric_thresholds=DataQualityMetricThreshold._from_rest_object(obj.metric_thresholds),
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
            properties=obj.properties,
        )

    @classmethod
    def _get_default_data_quality_signal(
        cls,
    ) -> "DataQualitySignal":
        return cls(
            features=ALL_FEATURES,  # type: ignore[arg-type]
            metric_thresholds=DataQualityMetricThreshold._get_default_thresholds(),
        )


@experimental
class FADProductionData(RestTranslatableMixin):
    """Feature Attribution Production Data

    :keyword input_data: Input data used by the monitor.
    :paramtype input_data: ~azure.ai.ml.Input
    :keyword data_context: The context of the input dataset. Accepted values are "model_inputs",
        "model_outputs", "training", "test", "validation", and "ground_truth".
    :paramtype data_context: ~azure.ai.ml.constants._monitoring
    :keyword data_column_names: The names of the columns in the input data.
    :paramtype data_column_names: Dict[str, str]
    :keyword pre_processing_component: The ARM (Azure Resource Manager) resource ID of the component resource used to
        preprocess the data.
    :paramtype pre_processing_component: string
    :param data_window: The number of days or a time frame that a singal monitor looks back over the target.
    :type data_window: BaselineDataRange
    """

    def __init__(
        self,
        *,
        input_data: Input,
        data_context: Optional[MonitorDatasetContext] = None,
        data_column_names: Optional[Dict[str, str]] = None,
        pre_processing_component: Optional[str] = None,
        data_window: Optional[BaselineDataRange] = None,
    ):
        self.input_data = input_data
        self.data_context = data_context
        self.data_column_names = data_column_names
        self.pre_processing_component = pre_processing_component
        self.data_window = data_window

    def _to_rest_object(self, **kwargs: Any) -> RestMonitoringInputData:
        default_data_window_size = kwargs.get("default")
        if self.data_window is None:
            self.data_window = BaselineDataRange(
                lookback_window_size=default_data_window_size, lookback_window_offset="P0D"
            )
        if self.data_window.lookback_window_size == "default":
            self.data_window.lookback_window_size = default_data_window_size
        uri = self.input_data.path
        job_type = self.input_data.type
        monitoring_input_data = TrailingInputData(
            data_context=self.data_context,
            target_columns=self.data_column_names,
            job_type=job_type,
            uri=uri,
            pre_processing_component_id=self.pre_processing_component,
            window_size=self.data_window.lookback_window_size,
            window_offset=(
                self.data_window.lookback_window_offset
                if self.data_window.lookback_window_offset is not None
                else "P0D"
            ),
        )
        return monitoring_input_data._to_rest_object()

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringInputData) -> "FADProductionData":
        data_window = BaselineDataRange(
            lookback_window_size=isodate.duration_isoformat(obj.window_size),
            lookback_window_offset=isodate.duration_isoformat(obj.window_offset),
        )
        return cls(
            input_data=Input(
                path=obj.uri,
                type=obj.job_input_type,
            ),
            data_context=obj.data_context,
            data_column_names=obj.columns,
            pre_processing_component=obj.preprocessing_component_id,
            data_window=data_window,
        )


@experimental
class FeatureAttributionDriftSignal(RestTranslatableMixin):
    """Feature attribution drift signal

    :ivar type: The type of the signal. Set to "feature_attribution_drift" for this class.
    :vartype type: str
    :keyword production_data: The data for which drift will be calculated.
    :paratype production_data: ~azure.ai.ml.entities.FADProductionData
    :keyword reference_data: The data to calculate drift against.
    :paramtype reference_data: ~azure.ai.ml.entities.ReferenceData
    :keyword metric_thresholds: Metrics to calculate and their
        associated thresholds.
    :paramtype metric_thresholds: ~azure.ai.ml.entities.FeatureAttributionDriftMetricThreshold
    :keyword alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    """

    def __init__(
        self,
        *,
        production_data: Optional[List[FADProductionData]] = None,
        reference_data: ReferenceData,
        metric_thresholds: FeatureAttributionDriftMetricThreshold,
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ):
        self.production_data = production_data
        self.reference_data = reference_data
        self.metric_thresholds = metric_thresholds
        self.alert_enabled = alert_enabled
        self.properties = properties
        self.type = MonitorSignalType.FEATURE_ATTRIBUTION_DRIFT

    def _to_rest_object(self, **kwargs: Any) -> RestFeatureAttributionDriftMonitoringSignal:
        default_window_size = kwargs.get("default_data_window_size")
        ref_data_window_size = kwargs.get("ref_data_window_size")
        return RestFeatureAttributionDriftMonitoringSignal(
            production_data=(
                [data._to_rest_object(default=default_window_size) for data in self.production_data]
                if self.production_data is not None
                else None
            ),
            reference_data=self.reference_data._to_rest_object(
                default_data_window=default_window_size, ref_data_window_size=ref_data_window_size
            ),
            metric_threshold=self.metric_thresholds._to_rest_object(),
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            properties=self.properties,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestFeatureAttributionDriftMonitoringSignal) -> "FeatureAttributionDriftSignal":
        return cls(
            production_data=[FADProductionData._from_rest_object(data) for data in obj.production_data],
            reference_data=ReferenceData._from_rest_object(obj.reference_data),
            metric_thresholds=FeatureAttributionDriftMetricThreshold._from_rest_object(obj.metric_threshold),
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
            properties=obj.properties,
        )


@experimental
class ModelPerformanceSignal(RestTranslatableMixin):
    """Model performance signal.

    :keyword baseline_dataset: The data to calculate performance against.
    :paramtype baseline_dataset: ~azure.ai.ml.entities.MonitorInputData
    :keyword metric_thresholds: A list of metrics to calculate and their
        associated thresholds.
    :paramtype metric_thresholds: ~azure.ai.ml.entities.ModelPerformanceMetricThreshold
    :keyword model_type: The model type.
    :paramtype model_type: ~azure.ai.ml.constants.MonitorModelType
    :keyword data_segment: The data segment to calculate performance against.
    :paramtype data_segment: ~azure.ai.ml.entities.DataSegment
    :keyword alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    """

    def __init__(
        self,
        *,
        production_data: ProductionData,
        reference_data: ReferenceData,
        metric_thresholds: ModelPerformanceMetricThreshold,
        data_segment: Optional[DataSegment] = None,
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ) -> None:
        self.production_data = production_data
        self.reference_data = reference_data
        self.metric_thresholds = metric_thresholds
        self.alert_enabled = alert_enabled
        self.type = MonitorSignalType.MODEL_PERFORMANCE
        self.data_segment = data_segment
        self.properties = properties

    def _to_rest_object(self, **kwargs: Any) -> RestModelPerformanceSignal:
        default_data_window_size = kwargs.get("default_data_window_size")
        ref_data_window_size = kwargs.get("ref_data_window_size")
        if self.properties is None:
            self.properties = {}
        self.properties["azureml.modelmonitor.model_performance_thresholds"] = self.metric_thresholds._to_str_object()
        if self.production_data.data_window is None:
            self.production_data.data_window = BaselineDataRange(
                lookback_window_size=default_data_window_size,
            )
        return RestModelPerformanceSignal(
            production_data=[self.production_data._to_rest_object(default_data_window_size=default_data_window_size)],
            reference_data=self.reference_data._to_rest_object(
                default_data_window_size=default_data_window_size, ref_data_window_size=ref_data_window_size
            ),
            metric_threshold=self.metric_thresholds._to_rest_object(),
            data_segment=self.data_segment._to_rest_object() if self.data_segment else None,
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            properties=self.properties,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestModelPerformanceSignal) -> "ModelPerformanceSignal":
        return cls(
            production_data=ProductionData._from_rest_object(obj.production_data[0]),
            reference_data=ReferenceData._from_rest_object(obj.reference_data),
            metric_thresholds=ModelPerformanceMetricThreshold._from_rest_object(obj.metric_threshold),
            data_segment=DataSegment._from_rest_object(obj.data_segment) if obj.data_segment else None,
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
        )


@experimental
class Connection(RestTranslatableMixin):
    """Monitoring Connection

    :param environment_variables: A dictionary of environment variables to set for the workspace.
    :paramtype environment_variables: Optional[dict[str, str]]
    :param secret_config: A dictionary of secrets to set for the workspace.
    :paramtype secret_config: Optional[dict[str, str]]
    """

    def __init__(
        self,
        *,
        environment_variables: Optional[Dict[str, str]] = None,
        secret_config: Optional[Dict[str, str]] = None,
    ):
        self.environment_variables = environment_variables
        self.secret_config = secret_config

    def _to_rest_object(self) -> RestMonitoringWorkspaceConnection:
        return RestMonitoringWorkspaceConnection(
            environment_variables=self.environment_variables,
            secrets=self.secret_config,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringWorkspaceConnection) -> "Connection":
        return cls(
            environment_variables=obj.environment_variables,
            secret_config=obj.secrets,
        )


@experimental
class CustomMonitoringSignal(RestTranslatableMixin):
    """Custom monitoring signal.

    :ivar type: The type of the signal. Set to "custom" for this class.
    :vartype type: str
    :keyword input_data: A dictionary of input datasets for monitoring.
        Each key is the component input port name, and its value is the data asset.
    :paramtype input_data: Optional[dict[str, ~azure.ai.ml.entities.ReferenceData]]
    :keyword metric_thresholds: A list of metrics to calculate and their
        associated thresholds.
    :paramtype metric_thresholds: List[~azure.ai.ml.entities.CustomMonitoringMetricThreshold]
    :keyword inputs:
    :paramtype inputs: Optional[dict[str, ~azure.ai.ml.entities.Input]]
    :keyword component_id: The ARM (Azure Resource Manager) ID of the component resource used to
        calculate the custom metrics.
    :paramtype component_id: str
    :keyword connection: Specify connection with environment variables and secret configs.
    :paramtype connection: Optional[~azure.ai.ml.entities.WorkspaceConnection]
    :keyword alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    :keyword properties: A dictionary of custom properties for the signal.
    :paramtype properties: Optional[dict[str, str]]
    """

    def __init__(
        self,
        *,
        inputs: Optional[Dict[str, Input]] = None,
        metric_thresholds: List[CustomMonitoringMetricThreshold],
        component_id: str,
        connection: Optional[Connection] = None,
        input_data: Optional[Dict[str, ReferenceData]] = None,
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
    ):
        self.type = MonitorSignalType.CUSTOM
        self.inputs = inputs
        self.metric_thresholds = metric_thresholds
        self.component_id = component_id
        self.alert_enabled = alert_enabled
        self.input_data = input_data
        self.properties = properties
        self.connection = connection

    def _to_rest_object(self, **kwargs: Any) -> RestCustomMonitoringSignal:  # pylint:disable=unused-argument
        if self.connection is None:
            self.connection = Connection()
        return RestCustomMonitoringSignal(
            component_id=self.component_id,
            metric_thresholds=[threshold._to_rest_object() for threshold in self.metric_thresholds],
            inputs=to_rest_dataset_literal_inputs(self.inputs, job_type=None) if self.inputs else None,
            input_assets=(
                {asset_name: asset_value._to_rest_object() for asset_name, asset_value in self.input_data.items()}
                if self.input_data
                else None
            ),
            workspace_connection=self.connection._to_rest_object(),
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            properties=self.properties,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestCustomMonitoringSignal) -> "CustomMonitoringSignal":
        return cls(
            inputs=from_rest_inputs_to_dataset_literal(obj.inputs) if obj.inputs else None,
            input_data={key: ReferenceData._from_rest_object(data) for key, data in obj.input_assets.items()},
            metric_thresholds=[
                CustomMonitoringMetricThreshold._from_rest_object(metric) for metric in obj.metric_thresholds
            ],
            component_id=obj.component_id,
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
            properties=obj.properties,
            connection=Connection._from_rest_object(obj.workspace_connection),
        )


@experimental
class LlmData(RestTranslatableMixin):
    """LLM Request Response Data

    :param input_data: Input data used by the monitor.
    :paramtype input_data: ~azure.ai.ml.entities.Input
    :param data_column_names: The names of columns in the input data.
    :paramtype data_column_names: Dict[str, str]
    :param data_window: The number of days or a time frame that a singal monitor looks back over the target.
    :type data_window_size: BaselineDataRange
    """

    def __init__(
        self,
        *,
        input_data: Input,
        data_column_names: Optional[Dict[str, str]] = None,
        data_window: Optional[BaselineDataRange] = None,
    ):
        self.input_data = input_data
        self.data_column_names = data_column_names
        self.data_window = data_window

    def _to_rest_object(self, **kwargs: Any) -> RestMonitoringInputData:
        if self.data_window is None:
            self.data_window = BaselineDataRange(
                lookback_window_size=kwargs.get("default"),
            )
        return TrailingInputData(
            target_columns=self.data_column_names,
            job_type=self.input_data.type,
            uri=self.input_data.path,
            window_size=self.data_window.lookback_window_size,
            window_offset=(
                self.data_window.lookback_window_offset
                if self.data_window.lookback_window_offset is not None
                else "P0D"
            ),
        )._to_rest_object()

    @classmethod
    def _from_rest_object(cls, obj: RestMonitoringInputData) -> "LlmData":
        data_window = BaselineDataRange(
            lookback_window_size=isodate.duration_isoformat(obj.window_size),
            lookback_window_offset=isodate.duration_isoformat(obj.window_offset),
        )
        return cls(
            input_data=Input(
                path=obj.uri,
                type=obj.job_input_type,
            ),
            data_column_names=obj.columns,
            data_window=data_window,
        )


@experimental
class GenerationSafetyQualitySignal(RestTranslatableMixin):
    """Generation Safety Quality monitoring signal.

    :ivar type: The type of the signal. Set to "generationsafetyquality" for this class.
    :vartype type: str
    :keyword production_data: A list of input datasets for monitoring.
    :paramtype input_datasets: Optional[dict[str, ~azure.ai.ml.entities.LlmData]]
    :keyword metric_thresholds: Metrics to calculate and their associated thresholds.
    :paramtype metric_thresholds: ~azure.ai.ml.entities.GenerationSafetyQualityMonitoringMetricThreshold
    :keyword alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    :keyword connection_id: Gets or sets the connection ID used to connect to the
        content generation endpoint.
    :paramtype connection_id: str
    :keyword properties: The properties of the signal
    :paramtype properties: Dict[str, str]
    :keyword sampling_rate: The sample rate of the target data, should be greater
        than 0 and at most 1.
    :paramtype sampling_rate: float
    """

    def __init__(
        self,
        *,
        production_data: Optional[List[LlmData]] = None,
        connection_id: Optional[str] = None,
        metric_thresholds: GenerationSafetyQualityMonitoringMetricThreshold,
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
        sampling_rate: Optional[float] = None,
    ):
        self.type = MonitorSignalType.GENERATION_SAFETY_QUALITY
        self.production_data = production_data
        self.connection_id = connection_id
        self.metric_thresholds = metric_thresholds
        self.alert_enabled = alert_enabled
        self.properties = properties
        self.sampling_rate = sampling_rate

    def _to_rest_object(self, **kwargs: Any) -> RestGenerationSafetyQualityMonitoringSignal:
        data_window_size = kwargs.get("default_data_window_size")
        return RestGenerationSafetyQualityMonitoringSignal(
            production_data=(
                [data._to_rest_object(default=data_window_size) for data in self.production_data]
                if self.production_data is not None
                else None
            ),
            workspace_connection_id=self.connection_id,
            metric_thresholds=self.metric_thresholds._to_rest_object(),
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            properties=self.properties,
            sampling_rate=self.sampling_rate,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestGenerationSafetyQualityMonitoringSignal) -> "GenerationSafetyQualitySignal":
        return cls(
            production_data=[LlmData._from_rest_object(data) for data in obj.production_data],
            connection_id=obj.workspace_connection_id,
            metric_thresholds=GenerationSafetyQualityMonitoringMetricThreshold._from_rest_object(obj.metric_thresholds),
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
            properties=obj.properties,
            sampling_rate=obj.sampling_rate,
        )


@experimental
class GenerationTokenStatisticsSignal(RestTranslatableMixin):
    """Generation token statistics signal definition.

    :ivar type: The type of the signal. Set to "generationtokenstatisticssignal" for this class.
    :vartype type: str
    :keyword production_data: input dataset for monitoring.
    :paramtype input_dataset: Optional[~azure.ai.ml.entities.LlmData]
    :keyword metric_thresholds: Metrics to calculate and their associated thresholds. Defaults to App Traces
    :paramtype metric_thresholds: Optional[~azure.ai.ml.entities.GenerationTokenStatisticsMonitorMetricThreshold]
    :keyword alert_enabled: Whether or not to enable alerts for the signal. Defaults to False.
    :paramtype alert_enabled: bool
    :keyword properties: The properties of the signal
    :paramtype properties: Optional[Dict[str, str]]
    :keyword sampling_rate: The sample rate of the target data, should be greater
        than 0 and at most 1.
    :paramtype sampling_rate: float

    .. admonition:: Example:

        .. literalinclude:: ../samples/ml_samples_genAI_monitors_configuration.py
                :start-after: [START default_monitoring]
                :end-before: [END default_monitoring]
                :language: python
                :dedent: 8
                :caption: Set Token Statistics Monitor.
    """

    def __init__(
        self,
        *,
        production_data: Optional[LlmData] = None,
        metric_thresholds: Optional[GenerationTokenStatisticsMonitorMetricThreshold] = None,
        alert_enabled: bool = False,
        properties: Optional[Dict[str, str]] = None,
        sampling_rate: Optional[float] = None,
    ):
        self.type = MonitorSignalType.GENERATION_TOKEN_STATISTICS
        self.production_data = production_data
        self.metric_thresholds = metric_thresholds
        self.alert_enabled = alert_enabled
        self.properties = properties
        self.sampling_rate = sampling_rate

    def _to_rest_object(self, **kwargs: Any) -> RestGenerationTokenStatisticsSignal:
        data_window_size = kwargs.get("default_data_window_size")
        return RestGenerationTokenStatisticsSignal(
            production_data=(
                self.production_data._to_rest_object(default=data_window_size)
                if self.production_data is not None
                else None
            ),
            metric_thresholds=(
                self.metric_thresholds._to_rest_object()
                if self.metric_thresholds
                else GenerationTokenStatisticsMonitorMetricThreshold._get_default_thresholds()._to_rest_object()
            ),
            mode=MonitoringNotificationMode.ENABLED if self.alert_enabled else MonitoringNotificationMode.DISABLED,
            properties=self.properties,
            sampling_rate=self.sampling_rate if self.sampling_rate else 0.1,
        )

    @classmethod
    def _from_rest_object(cls, obj: RestGenerationTokenStatisticsSignal) -> "GenerationTokenStatisticsSignal":
        return cls(
            production_data=LlmData._from_rest_object(obj.production_data),
            metric_thresholds=GenerationTokenStatisticsMonitorMetricThreshold._from_rest_object(obj.metric_thresholds),
            alert_enabled=(
                False
                if not obj.mode or (obj.mode and obj.mode == MonitoringNotificationMode.DISABLED)
                else MonitoringNotificationMode.ENABLED
            ),
            properties=obj.properties,
            sampling_rate=obj.sampling_rate,
        )

    @classmethod
    def _get_default_token_statistics_signal(cls) -> "GenerationTokenStatisticsSignal":
        return cls(
            metric_thresholds=GenerationTokenStatisticsMonitorMetricThreshold._get_default_thresholds(),
            sampling_rate=0.1,
        )


def _from_rest_features(
    obj: RestMonitoringFeatureFilterBase,
) -> Optional[Union[List[str], MonitorFeatureFilter, Literal["all_features"]]]:
    if isinstance(obj, RestTopNFeaturesByAttribution):
        return MonitorFeatureFilter(top_n_feature_importance=obj.top)
    if isinstance(obj, RestFeatureSubset):
        _restFeatureSubset: List[str] = obj.features
        return _restFeatureSubset
    if isinstance(obj, RestAllFeatures):
        _restAllFeatures: Literal["all_features"] = ALL_FEATURES  # type: ignore[assignment]
        return _restAllFeatures

    return None


def _to_rest_features(
    features: Union[List[str], MonitorFeatureFilter, Literal["all_features"]]
) -> RestMonitoringFeatureFilterBase:
    rest_features = None
    if isinstance(features, list):
        rest_features = RestFeatureSubset(features=features)
    elif isinstance(features, MonitorFeatureFilter):
        rest_features = features._to_rest_object()
    elif isinstance(features, str) and features == ALL_FEATURES:
        rest_features = RestAllFeatures()
    return rest_features


def _to_rest_num_cat_metrics(numerical_metrics: Any, categorical_metrics: Any) -> List:
    metrics = []
    if numerical_metrics is not None:
        metrics.append(numerical_metrics._to_rest_object())

    if categorical_metrics is not None:
        metrics.append(categorical_metrics._to_rest_object())

    return metrics


def _to_rest_data_quality_metrics(numerical_metrics: Any, categorical_metrics: Any) -> List:
    metric_thresholds: List = []
    if numerical_metrics is not None:
        metric_thresholds = metric_thresholds + numerical_metrics._to_rest_object()

    if categorical_metrics is not None:
        metric_thresholds = metric_thresholds + categorical_metrics._to_rest_object()

    return metric_thresholds



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""This file stores constants that will be used in mldesigner package."""
from azure.ai.ml._internal._schema.component import NodeType as V1NodeType
from azure.ai.ml._internal.entities import (
    Ae365exepool,
    AetherBridge,
    Command as InternalCommand,
    Parallel as InternalParallel,
    Pipeline as InternalPipeline,
    DataTransfer,
    Distributed,
    HDInsight,
    Hemera,
    Scope,
    Starlite,
)

V1_COMPONENT_TO_NODE = {
    V1NodeType.SCOPE: Scope,
    V1NodeType.COMMAND: InternalCommand,
    V1NodeType.PARALLEL: InternalParallel,
    V1NodeType.PIPELINE: InternalPipeline,
    V1NodeType.DATA_TRANSFER: DataTransfer,
    V1NodeType.DISTRIBUTED: Distributed,
    V1NodeType.HDI: HDInsight,
    V1NodeType.STARLITE: Starlite,
    V1NodeType.HEMERA: Hemera,
    V1NodeType.AE365EXEPOOL: Ae365exepool,
    V1NodeType.AETHER_BRIDGE: AetherBridge,
}



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Dict, Optional


class EndpointConnection:
    """Private Endpoint Connection related to a workspace private endpoint.

    :param subscription_id: Subscription id of the connection.
    :type subscription_id: str
    :param resource_group: Resource group of the connection.
    :type resource_group: str
    :param vnet_name: Name of the virtual network of the connection.
    :type vnet_name: str
    :param subnet_name: Name of the subnet of the connection.
    :type subnet_name: str
    :param location: Location of the connection.
    :type location: str
    """

    def __init__(
        self,
        subscription_id: str,
        resource_group: str,
        vnet_name: str,
        subnet_name: str,
        location: Optional[str] = None,
    ):
        self.subscription_id = subscription_id
        self.resource_group = resource_group
        self.location = location
        self.vnet_name = vnet_name
        self.subnet_name = subnet_name


class PrivateEndpoint:
    """Private Endpoint of a workspace.

    :param approval_type: Approval type of the private endpoint.
    :type approval_type: str
    :param connections: List of private endpoint connections.
    :type connections: List[~azure.ai.ml.entities.EndpointConnection]
    """

    def __init__(
        self,
        approval_type: Optional[str] = None,
        connections: Optional[Dict[str, EndpointConnection]] = None,
    ):
        self.approval_type = approval_type
        self.connections = connections



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import logging
import re
from os import PathLike
from typing import Any, Optional, Tuple, Union

from typing_extensions import Protocol

from azure.ai.ml._artifacts._artifact_utilities import _check_and_upload_env_build_context, _check_and_upload_path
from azure.ai.ml._scope_dependent_operations import (
    OperationConfig,
    OperationsContainer,
    OperationScope,
    _ScopeDependentOperations,
)
from azure.ai.ml._utils._arm_id_utils import (
    AMLLabelledArmId,
    AMLNamedArmId,
    AMLVersionedArmId,
    get_arm_id_with_version,
    is_ARM_id_for_resource,
    is_registry_id_for_resource,
    is_singularity_full_name_for_resource,
    is_singularity_id_for_resource,
    is_singularity_short_name_for_resource,
    parse_name_label,
    parse_prefixed_name_version,
)
from azure.ai.ml._utils._asset_utils import _resolve_label_to_asset, get_storage_info_for_non_registry_asset
from azure.ai.ml._utils._storage_utils import AzureMLDatastorePathUri
from azure.ai.ml.constants._common import (
    ARM_ID_PREFIX,
    AZUREML_RESOURCE_PROVIDER,
    CURATED_ENV_PREFIX,
    DEFAULT_LABEL_NAME,
    FILE_PREFIX,
    FOLDER_PREFIX,
    HTTPS_PREFIX,
    JOB_URI_REGEX_FORMAT,
    LABELLED_RESOURCE_ID_FORMAT,
    LABELLED_RESOURCE_NAME,
    MLFLOW_URI_REGEX_FORMAT,
    NAMED_RESOURCE_ID_FORMAT,
    REGISTRY_VERSION_PATTERN,
    SINGULARITY_FULL_NAME_REGEX_FORMAT,
    SINGULARITY_ID_FORMAT,
    SINGULARITY_SHORT_NAME_REGEX_FORMAT,
    VERSIONED_RESOURCE_ID_FORMAT,
    VERSIONED_RESOURCE_NAME,
    AzureMLResourceType,
)
from azure.ai.ml.entities import Component
from azure.ai.ml.entities._assets import Code, Data, Environment, Model
from azure.ai.ml.entities._assets.asset import Asset
from azure.ai.ml.exceptions import (
    AssetException,
    EmptyDirectoryError,
    ErrorCategory,
    ErrorTarget,
    MlException,
    ModelException,
    ValidationErrorType,
    ValidationException,
)
from azure.core.exceptions import HttpResponseError, ResourceNotFoundError

module_logger = logging.getLogger(__name__)


class OperationOrchestrator(object):
    def __init__(
        self,
        operation_container: OperationsContainer,
        operation_scope: OperationScope,
        operation_config: OperationConfig,
    ):
        self._operation_container = operation_container
        self._operation_scope = operation_scope
        self._operation_config = operation_config

    @property
    def _datastore_operation(self) -> _ScopeDependentOperations:
        return self._operation_container.all_operations[AzureMLResourceType.DATASTORE]

    @property
    def _code_assets(self) -> _ScopeDependentOperations:
        return self._operation_container.all_operations[AzureMLResourceType.CODE]

    @property
    def _model(self) -> _ScopeDependentOperations:
        return self._operation_container.all_operations[AzureMLResourceType.MODEL]

    @property
    def _environments(self) -> _ScopeDependentOperations:
        return self._operation_container.all_operations[AzureMLResourceType.ENVIRONMENT]

    @property
    def _data(self) -> _ScopeDependentOperations:
        return self._operation_container.all_operations[AzureMLResourceType.DATA]

    @property
    def _component(self) -> _ScopeDependentOperations:
        return self._operation_container.all_operations[AzureMLResourceType.COMPONENT]

    @property
    def _virtual_cluster(self) -> _ScopeDependentOperations:
        return self._operation_container.all_operations[AzureMLResourceType.VIRTUALCLUSTER]

    def get_asset_arm_id(
        self,
        asset: Optional[Union[str, Asset]],
        azureml_type: str,
        register_asset: bool = True,
        sub_workspace_resource: bool = True,
    ) -> Optional[Union[str, Asset]]:
        """This method converts AzureML Id to ARM Id. Or if the given asset is entity object, it tries to
        register/upload the asset based on register_asset and azureml_type.

        :param asset: The asset to resolve/register. It can be a ARM id or a entity's object.
        :type asset: Optional[Union[str, Asset]]
        :param azureml_type: The AzureML resource type. Defined in AzureMLResourceType.
        :type azureml_type: str
        :param register_asset: Indicates if the asset should be registered, defaults to True.
        :type register_asset: Optional[bool]
        :param sub_workspace_resource:
        :type sub_workspace_resource: Optional[bool]
        :raises ~azure.ai.ml.exceptions.ValidationException: Raised if asset's ID cannot be converted
            or asset cannot be successfully registered.
        :return: The ARM Id or entity object
        :rtype: Optional[Union[str, ~azure.ai.ml.entities.Asset]]
        """
        # pylint: disable=too-many-return-statements, too-many-branches
        if (
            asset is None
            or is_ARM_id_for_resource(asset, azureml_type, sub_workspace_resource)
            or is_registry_id_for_resource(asset)
            or is_singularity_id_for_resource(asset)
        ):
            return asset
        if is_singularity_full_name_for_resource(asset):
            return self._get_singularity_arm_id_from_full_name(str(asset))
        if is_singularity_short_name_for_resource(asset):
            return self._get_singularity_arm_id_from_short_name(str(asset))
        if isinstance(asset, str):
            if azureml_type in AzureMLResourceType.NAMED_TYPES:
                return NAMED_RESOURCE_ID_FORMAT.format(
                    self._operation_scope.subscription_id,
                    self._operation_scope.resource_group_name,
                    AZUREML_RESOURCE_PROVIDER,
                    self._operation_scope.workspace_name,
                    azureml_type,
                    asset,
                )
            if azureml_type in AzureMLResourceType.VERSIONED_TYPES:
                # Short form of curated env will be expanded on the backend side.
                # CLI strips off azureml: in the schema, appending it back as required by backend
                if azureml_type == AzureMLResourceType.ENVIRONMENT:
                    azureml_prefix = "azureml:"
                    # return the same value if resolved result is passed in
                    _asset = asset[len(azureml_prefix) :] if asset.startswith(azureml_prefix) else asset
                    if _asset.startswith(CURATED_ENV_PREFIX) or re.match(
                        REGISTRY_VERSION_PATTERN, f"{azureml_prefix}{_asset}"
                    ):
                        return f"{azureml_prefix}{_asset}"

                name, label = parse_name_label(asset)
                # TODO: remove this condition after label is fully supported for all versioned resources
                if label == DEFAULT_LABEL_NAME and azureml_type == AzureMLResourceType.COMPONENT:
                    return LABELLED_RESOURCE_ID_FORMAT.format(
                        self._operation_scope.subscription_id,
                        self._operation_scope.resource_group_name,
                        AZUREML_RESOURCE_PROVIDER,
                        self._operation_scope.workspace_name,
                        azureml_type,
                        name,
                        label,
                    )
                name, version = self._resolve_name_version_from_name_label(asset, azureml_type)
                if not version:
                    name, version = parse_prefixed_name_version(asset)

                if not version:
                    msg = (
                        "Failed to extract version when parsing asset {} of type {} as arm id. "
                        "Version must be provided."
                    )
                    raise ValidationException(
                        message=msg.format(asset, azureml_type),
                        target=ErrorTarget.ASSET,
                        no_personal_data_message=msg.format("", azureml_type),
                        error_category=ErrorCategory.USER_ERROR,
                        error_type=ValidationErrorType.MISSING_FIELD,
                    )
                if self._operation_scope.registry_name:
                    # Short form for env not supported with registry flow except when it's a curated env.
                    # Adding a graceful error message for the scenario
                    if not asset.startswith(CURATED_ENV_PREFIX):
                        msg = (
                            "Use fully qualified name to reference custom environments "
                            "when creating assets in registry. "
                            "The syntax for fully qualified names is "
                            "azureml://registries/azureml/environments/{{env-name}}/versions/{{version}}"
                        )
                        raise ValidationException(
                            message=msg.format(asset, azureml_type),
                            target=ErrorTarget.ASSET,
                            no_personal_data_message=msg.format("", azureml_type),
                            error_category=ErrorCategory.USER_ERROR,
                            error_type=ValidationErrorType.INVALID_VALUE,
                        )

                return VERSIONED_RESOURCE_ID_FORMAT.format(
                    self._operation_scope.subscription_id,
                    self._operation_scope.resource_group_name,
                    AZUREML_RESOURCE_PROVIDER,
                    self._operation_scope.workspace_name,
                    azureml_type,
                    name,
                    version,
                )
            msg = "Unsupported azureml type {} for asset: {}"
            raise ValidationException(
                message=msg.format(azureml_type, asset),
                target=ErrorTarget.ASSET,
                no_personal_data_message=msg.format(azureml_type, ""),
                error_type=ValidationErrorType.INVALID_VALUE,
            )
        if isinstance(asset, Asset):
            try:
                result: Any = None
                # TODO: once the asset redesign is finished, this logic can be replaced with unified API
                if azureml_type == AzureMLResourceType.CODE and isinstance(asset, Code):
                    result = self._get_code_asset_arm_id(asset, register_asset=register_asset)
                elif azureml_type == AzureMLResourceType.ENVIRONMENT and isinstance(asset, Environment):
                    result = self._get_environment_arm_id(asset, register_asset=register_asset)
                elif azureml_type == AzureMLResourceType.MODEL and isinstance(asset, Model):
                    result = self._get_model_arm_id(asset, register_asset=register_asset)
                elif azureml_type == AzureMLResourceType.DATA and isinstance(asset, Data):
                    result = self._get_data_arm_id(asset, register_asset=register_asset)
                elif azureml_type == AzureMLResourceType.COMPONENT and isinstance(asset, Component):
                    result = self._get_component_arm_id(asset)
                else:
                    msg = "Unsupported azureml type {} for asset: {}"
                    raise ValidationException(
                        message=msg.format(azureml_type, asset),
                        target=ErrorTarget.ASSET,
                        no_personal_data_message=msg.format(azureml_type, ""),
                        error_category=ErrorCategory.USER_ERROR,
                        error_type=ValidationErrorType.INVALID_VALUE,
                    )
            except EmptyDirectoryError as e:
                msg = f"Error creating {azureml_type} asset : {e.message}"
                raise AssetException(
                    message=msg.format(azureml_type, e.message),
                    target=ErrorTarget.ASSET,
                    no_personal_data_message=msg.format(azureml_type, ""),
                    error=e,
                    error_category=ErrorCategory.SYSTEM_ERROR,
                ) from e
            return result
        msg = f"Error creating {azureml_type} asset: must be type Optional[Union[str, Asset]]"
        raise ValidationException(
            message=msg,
            target=ErrorTarget.ASSET,
            no_personal_data_message=msg,
            error_category=ErrorCategory.USER_ERROR,
            error_type=ValidationErrorType.INVALID_VALUE,
        )

    def _get_code_asset_arm_id(self, code_asset: Code, register_asset: bool = True) -> Union[Code, str]:
        try:
            self._validate_datastore_name(code_asset.path)
            if register_asset:
                code_asset = self._code_assets.create_or_update(code_asset)  # type: ignore[attr-defined]
                return str(code_asset.id)
            sas_info = get_storage_info_for_non_registry_asset(
                service_client=self._code_assets._service_client,  # type: ignore[attr-defined]
                workspace_name=self._operation_scope.workspace_name,
                name=code_asset.name,
                version=code_asset.version,
                resource_group=self._operation_scope.resource_group_name,
            )
            uploaded_code_asset, _ = _check_and_upload_path(
                artifact=code_asset,
                asset_operations=self._code_assets,  # type: ignore[arg-type]
                artifact_type=ErrorTarget.CODE,
                show_progress=self._operation_config.show_progress,
                sas_uri=sas_info["sas_uri"],
                blob_uri=sas_info["blob_uri"],
            )
            uploaded_code_asset._id = get_arm_id_with_version(
                self._operation_scope,
                AzureMLResourceType.CODE,
                code_asset.name,
                code_asset.version,
            )
            return uploaded_code_asset
        except (MlException, HttpResponseError) as e:
            raise e
        except Exception as e:
            raise AssetException(
                message=f"Error with code: {e}",
                target=ErrorTarget.ASSET,
                no_personal_data_message="Error getting code asset",
                error=e,
                error_category=ErrorCategory.SYSTEM_ERROR,
            ) from e

    def _get_environment_arm_id(self, environment: Environment, register_asset: bool = True) -> Union[str, Environment]:
        if register_asset:
            if environment.id:
                return environment.id
            env_response = self._environments.create_or_update(environment)  # type: ignore[attr-defined]
            return env_response.id
        environment = _check_and_upload_env_build_context(
            environment=environment,
            operations=self._environments,  # type: ignore[arg-type]
            show_progress=self._operation_config.show_progress,
        )
        environment._id = get_arm_id_with_version(
            self._operation_scope,
            AzureMLResourceType.ENVIRONMENT,
            environment.name,
            environment.version,
        )
        return environment

    def _get_model_arm_id(self, model: Model, register_asset: bool = True) -> Union[str, Model]:
        try:
            if not is_registry_id_for_resource(model.id):
                self._validate_datastore_name(model.path)

            if register_asset:
                if model.id:
                    return model.id
                return self._model.create_or_update(model).id  # type: ignore[attr-defined]
            uploaded_model, _ = _check_and_upload_path(
                artifact=model,
                asset_operations=self._model,  # type: ignore[arg-type]
                artifact_type=ErrorTarget.MODEL,
                show_progress=self._operation_config.show_progress,
            )
            uploaded_model._id = get_arm_id_with_version(
                self._operation_scope,
                AzureMLResourceType.MODEL,
                model.name,
                model.version,
            )
            return uploaded_model
        except (MlException, HttpResponseError) as e:
            raise e
        except Exception as e:
            raise ModelException(
                message=f"Error with model: {e}",
                target=ErrorTarget.MODEL,
                no_personal_data_message="Error getting model",
                error=e,
                error_category=ErrorCategory.SYSTEM_ERROR,
            ) from e

    def _get_data_arm_id(self, data_asset: Data, register_asset: bool = True) -> Union[str, Data]:
        self._validate_datastore_name(data_asset.path)

        if register_asset:
            return self._data.create_or_update(data_asset).id  # type: ignore[attr-defined]
        data_asset, _ = _check_and_upload_path(
            artifact=data_asset,
            asset_operations=self._data,  # type: ignore[arg-type]
            artifact_type=ErrorTarget.DATA,
            show_progress=self._operation_config.show_progress,
        )
        return data_asset

    def _get_component_arm_id(self, component: Component) -> str:
        """Gets the component ARM ID.

        :param component: The component
        :type component: Component
        :return: The component id
        :rtype: str
        """

        # If component arm id is already resolved, return the id otherwise get arm id via remote call.
        # Register the component if necessary, and FILL BACK the arm id to component to reduce remote call.
        if not component.id:
            component._id = self._component.create_or_update(  # type: ignore[attr-defined]
                component, is_anonymous=True, show_progress=self._operation_config.show_progress
            ).id
        return str(component.id)

    def _get_singularity_arm_id_from_full_name(self, singularity: str) -> str:
        match = re.match(SINGULARITY_FULL_NAME_REGEX_FORMAT, singularity)
        subscription_id = match.group("subscription_id") if match is not None else ""
        resource_group_name = match.group("resource_group_name") if match is not None else ""
        vc_name = match.group("name") if match is not None else ""
        arm_id = SINGULARITY_ID_FORMAT.format(subscription_id, resource_group_name, vc_name)
        vc = self._virtual_cluster.get(arm_id)  # type: ignore[attr-defined]
        return str(vc["id"])

    def _get_singularity_arm_id_from_short_name(self, singularity: str) -> str:
        match = re.match(SINGULARITY_SHORT_NAME_REGEX_FORMAT, singularity)
        vc_name = match.group("name") if match is not None else ""
        # below list operation can be time-consuming, may need an optimization on this
        match_vcs = [vc for vc in self._virtual_cluster.list() if vc["name"] == vc_name]  # type: ignore[attr-defined]
        num_match_vc = len(match_vcs)
        if num_match_vc != 1:
            if num_match_vc == 0:
                msg = "The virtual cluster {} could not be found."
            else:
                msg = "More than one match virtual clusters {} found."
            raise ValidationException(
                message=msg.format(vc_name),
                no_personal_data_message=msg.format(""),
                target=ErrorTarget.COMPUTE,
                error_type=ValidationErrorType.INVALID_VALUE,
            )
        return str(match_vcs[0]["id"])

    def _resolve_name_version_from_name_label(self, aml_id: str, azureml_type: str) -> Tuple[str, Optional[str]]:
        """Given an AzureML id of the form name@label, resolves the label to the actual ID.

        :param aml_id: AzureML id of the form name@label
        :type aml_id: str
        :param azureml_type: The AzureML resource type. Defined in AzureMLResourceType.
        :type azureml_type: str
        :return: Returns tuple (name, version) on success, (name@label, None) if resolution fails
        :rtype: Tuple[str, Optional[str]]
        """
        name, label = parse_name_label(aml_id)
        if (
            azureml_type not in AzureMLResourceType.VERSIONED_TYPES
            or azureml_type == AzureMLResourceType.CODE
            or not label
        ):
            return aml_id, None

        return (
            name,
            _resolve_label_to_asset(
                self._operation_container.all_operations[azureml_type],
                name,
                label=label,
            ).version,
        )

    # pylint: disable=unused-argument
    def resolve_azureml_id(self, arm_id: Optional[str] = None, **kwargs: Any) -> Optional[str]:
        """This function converts ARM id to name or name:version AzureML id. It parses the ARM id and matches the
        subscription Id, resource group name and workspace_name.

        TODO: It is debatable whether this method should be in operation_orchestrator.

        :param arm_id: entity's ARM id, defaults to None
        :type arm_id: str
        :return: AzureML id
        :rtype: str
        """

        if arm_id:
            if not isinstance(arm_id, str):
                msg = "arm_id cannot be resolved: str expected but got {}".format(type(arm_id))  # type: ignore
                raise ValidationException(
                    message=msg,
                    no_personal_data_message=msg,
                    target=ErrorTarget.GENERAL,
                    error_type=ValidationErrorType.INVALID_VALUE,
                )
            try:
                arm_id_obj = AMLVersionedArmId(arm_id)
                if arm_id_obj.is_registry_id:
                    return arm_id
                if self._match(arm_id_obj):
                    return str(VERSIONED_RESOURCE_NAME.format(arm_id_obj.asset_name, arm_id_obj.asset_version))
            except ValidationException:
                pass  # fall back to named arm id
            try:
                arm_id_obj = AMLLabelledArmId(arm_id)
                if self._match(arm_id_obj):
                    return str(LABELLED_RESOURCE_NAME.format(arm_id_obj.asset_name, arm_id_obj.asset_label))
            except ValidationException:
                pass  # fall back to named arm id
            try:
                arm_id_obj = AMLNamedArmId(arm_id)
                if self._match(arm_id_obj):
                    return str(arm_id_obj.asset_name)
            except ValidationException:
                pass  # fall back to be not a ARM_id
        return arm_id

    def _match(self, id_: Any) -> bool:
        return bool(
            (
                id_.subscription_id == self._operation_scope.subscription_id
                and id_.resource_group_name == self._operation_scope.resource_group_name
                and id_.workspace_name == self._operation_scope.workspace_name
            )
        )

    def _validate_datastore_name(self, datastore_uri: Optional[Union[str, PathLike]]) -> None:
        if datastore_uri:
            try:
                if isinstance(datastore_uri, str):
                    if datastore_uri.startswith(FILE_PREFIX):
                        datastore_uri = datastore_uri[len(FILE_PREFIX) :]
                    elif datastore_uri.startswith(FOLDER_PREFIX):
                        datastore_uri = datastore_uri[len(FOLDER_PREFIX) :]
                elif isinstance(datastore_uri, PathLike):
                    return

                if datastore_uri.startswith(HTTPS_PREFIX) and datastore_uri.count("/") == 7:
                    # only long-form (i.e. "https://x.blob.core.windows.net/datastore/LocalUpload/guid/x/x")
                    # format includes datastore
                    datastore_name = datastore_uri.split("/")[3]
                elif datastore_uri.startswith(ARM_ID_PREFIX) and not (
                    re.match(MLFLOW_URI_REGEX_FORMAT, datastore_uri) or re.match(JOB_URI_REGEX_FORMAT, datastore_uri)
                ):
                    datastore_name = AzureMLDatastorePathUri(datastore_uri).datastore
                else:
                    # local path
                    return

                if datastore_name.startswith(ARM_ID_PREFIX):
                    datastore_name = datastore_name[len(ARM_ID_PREFIX) :]

                self._datastore_operation.get(datastore_name)  # type: ignore[attr-defined]
            except ResourceNotFoundError as e:
                msg = "The datastore {} could not be found in this workspace."
                raise ValidationException(
                    message=msg.format(datastore_name),
                    target=ErrorTarget.DATASTORE,
                    no_personal_data_message=msg.format(""),
                    error_category=ErrorCategory.USER_ERROR,
                    error_type=ValidationErrorType.RESOURCE_NOT_FOUND,
                ) from e


class _AssetResolver(Protocol):
    """Describes the type of a function used by operation classes like :py:class:`JobOperations` and
    :py:class:`ComponentOperations` to resolve Assets

    .. see-also:: methods :py:method:`OperationOrchestrator.get_asset_arm_id`,
            :py:method:`OperationOrchestrator.resolve_azureml_id`

    """

    def __call__(
        self,
        asset: Optional[Union[str, Asset]],
        azureml_type: str,
        register_asset: bool = True,
        sub_workspace_resource: bool = True,
    ) -> Optional[Union[str, Asset]]:
        """Resolver function

        :param asset: The asset to resolve/register. It can be a ARM id or a entity's object.
        :type asset: Optional[Union[str, Asset]]
        :param azureml_type: The AzureML resource type. Defined in AzureMLResourceType.
        :type azureml_type: str
        :param register_asset: Indicates if the asset should be registered, defaults to True.
        :type register_asset: Optional[bool]
        :param sub_workspace_resource:
        :type sub_workspace_resource: Optional[bool]
        :raises ~azure.ai.ml.exceptions.ValidationException: Raised if asset's ID cannot be converted
            or asset cannot be successfully registered.
        :return: The ARM Id or entity object
        :rtype: Optional[Union[str, ~azure.ai.ml.entities.Asset]]
        """



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import List, Optional

from azure.ai.ml._restclient.v2023_02_01_preview.models import NotificationSetting as RestNotificationSetting
from azure.ai.ml.entities._mixins import RestTranslatableMixin


class Notification(RestTranslatableMixin):
    """Configuration for notification.

    :param email_on: Send email notification to user on specified notification type. Accepted values are
        "JobCompleted", "JobFailed", and "JobCancelled".
    :type email_on: Optional[list[str]]
    :param: The email recipient list which. Note that this parameter has a character limit of 499 which
        includes all of the recipient strings and each comma seperator.
    :paramtype emails: Optional[list[str]]
    """

    def __init__(self, *, email_on: Optional[List[str]] = None, emails: Optional[List[str]] = None) -> None:
        self.email_on = email_on
        self.emails = emails

    def _to_rest_object(self) -> RestNotificationSetting:
        return RestNotificationSetting(email_on=self.email_on, emails=self.emails)

    @classmethod
    def _from_rest_object(cls, obj: RestNotificationSetting) -> Optional["Notification"]:
        if not obj:
            return None
        return Notification(email_on=obj.email_on, emails=obj.emails)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

from typing import Any, Dict, Iterable, Optional, cast

from azure.ai.ml._restclient.v2024_04_01_preview import AzureMachineLearningWorkspaces as ServiceClient082023Preview
from azure.ai.ml._scope_dependent_operations import (
    OperationConfig,
    OperationsContainer,
    OperationScope,
    _ScopeDependentOperations,
)
from azure.ai.ml._telemetry import ActivityType, monitor_with_activity
from azure.ai.ml._utils._logger_utils import OpsLogger
from azure.ai.ml._utils.utils import _snake_to_camel
from azure.ai.ml.entities._credentials import ApiKeyConfiguration
from azure.ai.ml.entities._workspace.connections.workspace_connection import WorkspaceConnection
from azure.core.credentials import TokenCredential
from azure.core.tracing.decorator import distributed_trace

ops_logger = OpsLogger(__name__)
module_logger = ops_logger.module_logger


class WorkspaceConnectionsOperations(_ScopeDependentOperations):
    """WorkspaceConnectionsOperations.

    You should not instantiate this class directly. Instead, you should create
    an MLClient instance that instantiates it for you and attaches it as an attribute.
    """

    def __init__(
        self,
        operation_scope: OperationScope,
        operation_config: OperationConfig,
        service_client: ServiceClient082023Preview,
        all_operations: OperationsContainer,
        credentials: Optional[TokenCredential] = None,
        **kwargs: Dict,
    ):
        super(WorkspaceConnectionsOperations, self).__init__(operation_scope, operation_config)
        ops_logger.update_filter()
        self._all_operations = all_operations
        self._operation = service_client.workspace_connections
        self._credentials = credentials
        self._init_kwargs = kwargs

    def _try_fill_api_key(self, connection: WorkspaceConnection) -> None:
        """Try to fill in a connection's credentials with it's actual values.
        Connection data retrievals normally return an empty credential object that merely includes the
        connection's credential type, but not the actual secrets of that credential.
        However, it's extremely common for users to want to know the contents of their connection's credentials.
        This method tries to fill in the user's credentials with the actual values by making
        a secondary API call to the service. It requires that the user have the necessary permissions to do so,
        and it only works on api key-based credentials.

        :param connection: The connection to try to fill in the credentials for.
        :type connection: ~azure.ai.ml.entities.WorkspaceConnection
        """
        if hasattr(connection, "credentials") and isinstance(connection.credentials, ApiKeyConfiguration):
            list_secrets_response = self._operation.list_secrets(
                connection_name=connection.name,
                resource_group_name=self._operation_scope.resource_group_name,
                workspace_name=self._operation_scope.workspace_name,
            )
            if list_secrets_response.properties.credentials is not None:
                connection.credentials.key = list_secrets_response.properties.credentials.key

    @distributed_trace
    @monitor_with_activity(ops_logger, "WorkspaceConnections.Get", ActivityType.PUBLICAPI)
    def get(self, name: str, *, populate_secrets: bool = False, **kwargs: Dict) -> WorkspaceConnection:
        """Get a connection by name.

        :param name: Name of the connection.
        :type name: str
        :keyword populate_secrets: If true, make a secondary API call to try filling in the workspace
            connections credentials. Currently only works for api key-based credentials. Defaults to False.
        :paramtype populate_secrets: bool
        :raises ~azure.core.exceptions.HttpResponseError: Raised if the corresponding name and version cannot be
            retrieved from the service.
        :return: The connection with the provided name.
        :rtype: ~azure.ai.ml.entities.WorkspaceConnection
        """

        connection = WorkspaceConnection._from_rest_object(
            rest_obj=self._operation.get(
                workspace_name=self._workspace_name,
                connection_name=name,
                **self._scope_kwargs,
                **kwargs,
            )
        )

        if populate_secrets and connection is not None:
            self._try_fill_api_key(connection)
        return connection  # type: ignore[return-value]

    @distributed_trace
    @monitor_with_activity(ops_logger, "WorkspaceConnections.CreateOrUpdate", ActivityType.PUBLICAPI)
    def create_or_update(
        self, workspace_connection: WorkspaceConnection, *, populate_secrets: bool = False, **kwargs: Any
    ) -> WorkspaceConnection:
        """Create or update a connection.

        :param workspace_connection: Definition of a Workspace Connection or one of its subclasses
            or object which can be translated to a connection.
        :type workspace_connection: ~azure.ai.ml.entities.WorkspaceConnection
        :keyword populate_secrets: If true, make a secondary API call to try filling in the workspace
            connections credentials. Currently only works for api key-based credentials. Defaults to False.
        :paramtype populate_secrets: bool
        :return: Created or update connection.
        :rtype: ~azure.ai.ml.entities.WorkspaceConnection
        """
        rest_workspace_connection = workspace_connection._to_rest_object()
        response = self._operation.create(
            workspace_name=self._workspace_name,
            connection_name=workspace_connection.name,
            body=rest_workspace_connection,
            **self._scope_kwargs,
            **kwargs,
        )
        conn = WorkspaceConnection._from_rest_object(rest_obj=response)
        if populate_secrets and conn is not None:
            self._try_fill_api_key(conn)
        return conn

    @distributed_trace
    @monitor_with_activity(ops_logger, "WorkspaceConnections.Delete", ActivityType.PUBLICAPI)
    def delete(self, name: str, **kwargs: Any) -> None:
        """Delete the connection.

        :param name: Name of the connection.
        :type name: str
        """

        self._operation.delete(
            connection_name=name,
            workspace_name=self._workspace_name,
            **self._scope_kwargs,
            **kwargs,
        )

    @distributed_trace
    @monitor_with_activity(ops_logger, "WorkspaceConnections.List", ActivityType.PUBLICAPI)
    def list(
        self,
        connection_type: Optional[str] = None,
        *,
        populate_secrets: bool = False,
        include_data_connections: bool = False,
        **kwargs: Any,
    ) -> Iterable[WorkspaceConnection]:
        """List all connections for a workspace.

        :param connection_type: Type of connection to list.
        :type connection_type: Optional[str]
        :keyword populate_secrets: If true, make a secondary API call to try filling in the workspace
            connections credentials. Currently only works for api key-based credentials. Defaults to False.
        :paramtype populate_secrets: bool
        :keyword include_data_connections: If true, also return data connections. Defaults to False.
        :paramtype include_data_connections: bool
        :return: An iterator like instance of connection objects
        :rtype: Iterable[~azure.ai.ml.entities.WorkspaceConnection]
        """

        if include_data_connections:
            if "params" in kwargs:
                kwargs["params"]["includeAll"] = "true"
            else:
                kwargs["params"] = {"includeAll": "true"}

        def post_process_conn(rest_obj):
            result = WorkspaceConnection._from_rest_object(rest_obj)
            if populate_secrets and result is not None:
                self._try_fill_api_key(result)
            return result

        result = self._operation.list(
            workspace_name=self._workspace_name,
            cls=lambda objs: [post_process_conn(obj) for obj in objs],
            category=_snake_to_camel(connection_type) if connection_type else connection_type,
            **self._scope_kwargs,
            **kwargs,
        )

        return cast(Iterable[WorkspaceConnection], result)



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from ._operations import Operations
from ._workspaces_operations import WorkspacesOperations
from ._usages_operations import UsagesOperations
from ._virtual_machine_sizes_operations import VirtualMachineSizesOperations
from ._quotas_operations import QuotasOperations
from ._compute_operations import ComputeOperations
from ._private_endpoint_connections_operations import PrivateEndpointConnectionsOperations
from ._private_link_resources_operations import PrivateLinkResourcesOperations
from ._workspace_connections_operations import WorkspaceConnectionsOperations
from ._registries_operations import RegistriesOperations
from ._workspace_features_operations import WorkspaceFeaturesOperations
from ._registry_code_containers_operations import RegistryCodeContainersOperations
from ._registry_code_versions_operations import RegistryCodeVersionsOperations
from ._registry_component_containers_operations import RegistryComponentContainersOperations
from ._registry_component_versions_operations import RegistryComponentVersionsOperations
from ._registry_data_containers_operations import RegistryDataContainersOperations
from ._registry_data_versions_operations import RegistryDataVersionsOperations
from ._registry_environment_containers_operations import RegistryEnvironmentContainersOperations
from ._registry_environment_versions_operations import RegistryEnvironmentVersionsOperations
from ._registry_model_containers_operations import RegistryModelContainersOperations
from ._registry_model_versions_operations import RegistryModelVersionsOperations
from ._batch_endpoints_operations import BatchEndpointsOperations
from ._batch_deployments_operations import BatchDeploymentsOperations
from ._code_containers_operations import CodeContainersOperations
from ._code_versions_operations import CodeVersionsOperations
from ._component_containers_operations import ComponentContainersOperations
from ._component_versions_operations import ComponentVersionsOperations
from ._data_containers_operations import DataContainersOperations
from ._data_versions_operations import DataVersionsOperations
from ._datastores_operations import DatastoresOperations
from ._environment_containers_operations import EnvironmentContainersOperations
from ._environment_versions_operations import EnvironmentVersionsOperations
from ._featureset_containers_operations import FeaturesetContainersOperations
from ._featureset_versions_operations import FeaturesetVersionsOperations
from ._featurestore_entity_containers_operations import FeaturestoreEntityContainersOperations
from ._featurestore_entity_versions_operations import FeaturestoreEntityVersionsOperations
from ._jobs_operations import JobsOperations
from ._labeling_jobs_operations import LabelingJobsOperations
from ._model_containers_operations import ModelContainersOperations
from ._model_versions_operations import ModelVersionsOperations
from ._online_endpoints_operations import OnlineEndpointsOperations
from ._online_deployments_operations import OnlineDeploymentsOperations
from ._schedules_operations import SchedulesOperations

__all__ = [
    "Operations",
    "WorkspacesOperations",
    "UsagesOperations",
    "VirtualMachineSizesOperations",
    "QuotasOperations",
    "ComputeOperations",
    "PrivateEndpointConnectionsOperations",
    "PrivateLinkResourcesOperations",
    "WorkspaceConnectionsOperations",
    "RegistriesOperations",
    "WorkspaceFeaturesOperations",
    "RegistryCodeContainersOperations",
    "RegistryCodeVersionsOperations",
    "RegistryComponentContainersOperations",
    "RegistryComponentVersionsOperations",
    "RegistryDataContainersOperations",
    "RegistryDataVersionsOperations",
    "RegistryEnvironmentContainersOperations",
    "RegistryEnvironmentVersionsOperations",
    "RegistryModelContainersOperations",
    "RegistryModelVersionsOperations",
    "BatchEndpointsOperations",
    "BatchDeploymentsOperations",
    "CodeContainersOperations",
    "CodeVersionsOperations",
    "ComponentContainersOperations",
    "ComponentVersionsOperations",
    "DataContainersOperations",
    "DataVersionsOperations",
    "DatastoresOperations",
    "EnvironmentContainersOperations",
    "EnvironmentVersionsOperations",
    "FeaturesetContainersOperations",
    "FeaturesetVersionsOperations",
    "FeaturestoreEntityContainersOperations",
    "FeaturestoreEntityVersionsOperations",
    "JobsOperations",
    "LabelingJobsOperations",
    "ModelContainersOperations",
    "ModelVersionsOperations",
    "OnlineEndpointsOperations",
    "OnlineDeploymentsOperations",
    "SchedulesOperations",
]



import argparse
from operator import add

# from pyspark import SparkContext, SparkConf
from pyspark.sql import SparkSession

parser = argparse.ArgumentParser()
parser.add_argument("--input1")
args = parser.parse_args()

spark = SparkSession.builder.appName("PythonWordCount").getOrCreate()

lines = spark.read.text(args.input1).rdd.map(lambda r: r[0])
counts = lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
output = counts.collect()
for word, count in output:
    print("%s: %i" % (word, count))



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Optional

from azure.ai.ml._utils._arm_id_utils import is_ARM_id_for_resource
from azure.ai.ml._scope_dependent_operations import OperationScope

OPEN_AI_PROTOCOL_TEMPLATE = "azure_open_ai://deployment/{}/model/{}"


def build_open_ai_protocol(
    model: Optional[str] = None,
    deployment: Optional[str] = None,
):
    if not deployment or not model:
        return None
    return OPEN_AI_PROTOCOL_TEMPLATE.format(deployment, model)


def build_connection_id(id: Optional[str], scope: OperationScope):
    if not id or not scope.subscription_id or not scope.resource_group_name or not scope.workspace_name:
        return id

    if is_ARM_id_for_resource(id, "connections", True):
        return id

    # pylint: disable=line-too-long
    template = "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.MachineLearningServices/workspaces/{workspace_name}/connections/{id}"
    return template.format(
        subscription_id=scope.subscription_id,
        resource_group_name=scope.resource_group_name,
        workspace_name=scope.workspace_name,
        id=id,
    )



import random
import re
from typing import Callable

import pytest
from devtools_testutils import AzureRecordedTestCase
from test_utilities.utils import sleep_if_live

from azure.ai.ml import MLClient, load_environment
from azure.ai.ml._restclient.v2022_05_01.models import ListViewType
from azure.ai.ml.constants._common import ARM_ID_PREFIX
from azure.core.paging import ItemPaged


@pytest.fixture
def env_name(variable_recorder) -> Callable[[str], str]:
    def generate_random_environment_name(env_name: Callable[[str], str]) -> str:
        random_env_name = f"env-test-{str(random.randint(1, 10000000))}"
        return variable_recorder.get_or_record(env_name, random_env_name)

    return generate_random_environment_name


# previous bodiless_matcher fixture doesn't take effect because of typo, please add it in method level if needed


@pytest.mark.e2etest
@pytest.mark.usefixtures("recorded_test", "mock_code_hash")
@pytest.mark.production_experiences_test
class TestEnvironment(AzureRecordedTestCase):
    def test_environment_create_conda(self, client: MLClient, env_name: Callable[[str], str]) -> None:
        params_override = [{"name": env_name("name")}]
        env = load_environment(
            source="./tests/test_configs/environment/environment_conda.yml", params_override=params_override
        )

        environment = client._environments.create_or_update(env)

        environment_id = _get_environment_arm_id(
            client=client, environment_name=env.name, environment_version=env.version
        )

        assert environment.conda_file
        assert environment.id == environment_id

        env_dump = environment._to_dict()
        assert env_dump
        assert env_dump["id"] == ARM_ID_PREFIX + environment_id
        assert env_dump["conda_file"]
        assert env_dump["description"]

    def test_environment_create_conda_inline(self, client: MLClient, env_name: Callable[[str], str]) -> None:
        params_override = [{"name": env_name("name")}]
        env = load_environment(
            source="./tests/test_configs/environment/environment_conda_inline.yml", params_override=params_override
        )
        environment = client._environments.create_or_update(env)
        environment_id = _get_environment_arm_id(
            client=client, environment_name=env.name, environment_version=env.version
        )

        assert environment.conda_file
        assert environment.id == environment_id

        env_dump = environment._to_dict()
        assert env_dump
        assert env_dump["id"] == ARM_ID_PREFIX + environment_id
        assert env_dump["conda_file"]
        assert env_dump["description"]

    def test_environment_create_or_update_docker(self, client: MLClient, env_name: Callable[[str], str]) -> None:
        params_override = [{"name": env_name("name")}]
        env = load_environment(
            source="./tests/test_configs/environment/environment_docker_image.yml",
            params_override=params_override,
        )
        environment = client._environments.create_or_update(env)

        environment_id = _get_environment_arm_id(
            client=client, environment_name=env.name, environment_version=env.version
        )

        assert environment
        assert environment.image
        assert environment.id == environment_id

        env_dump = environment._to_dict()
        assert env_dump
        assert env_dump["id"] == ARM_ID_PREFIX + environment_id
        assert env_dump["image"] == environment.image

    @pytest.mark.live_test_only("Needs re-recording to work with new test proxy sanitizers")
    def test_environment_create_or_update_docker_context(
        self, client: MLClient, env_name: Callable[[str], str]
    ) -> None:
        params_override = [{"name": env_name("name")}]
        env = load_environment(
            source="./tests/test_configs/environment/environment_docker_context.yml",
            params_override=params_override,
        )
        environment = client._environments.create_or_update(env)

        environment_id = _get_environment_arm_id(
            client=client, environment_name=env.name, environment_version=env.version
        )

        assert environment
        assert environment.build
        assert environment.id == environment_id

        env_dump = environment._to_dict()
        assert env_dump
        assert env_dump["id"] == ARM_ID_PREFIX + environment_id
        assert env_dump["build"]
        assert env_dump["build"]["path"]
        assert env_dump["build"]["dockerfile_path"]

        context_uri = environment.build.path
        dockerfile_path = environment.build.dockerfile_path

        env.build.path = context_uri
        env.build.dockerfile_path = dockerfile_path

        environment_with_context_uri = client._environments.create_or_update(env)
        assert environment_with_context_uri
        assert environment_with_context_uri.build
        assert environment_with_context_uri.id == environment_id
        assert environment_with_context_uri.build.dockerfile_path == dockerfile_path
        assert environment_with_context_uri.build.path == context_uri

    def test_environment_create_or_update_docker_context_and_image(
        self, client: MLClient, env_name: Callable[[str], str]
    ) -> None:
        params_override = [{"name": env_name("name")}]
        env = load_environment(
            source="./tests/test_configs/environment/environment_docker_context.yml",
            params_override=params_override,
        )
        env.image = "blah"

        with pytest.raises(Exception) as error:
            client._environments.create_or_update(env)

        assert "Docker image or Dockerfile should be provided not both" in str(error.value)

    def test_environment_list(self, client: MLClient) -> None:
        environment_list = client._environments.list()
        assert environment_list
        assert isinstance(environment_list, ItemPaged)

    def test_environment_get(self, client: MLClient) -> None:
        environment_name = "AzureML-sklearn-1.0-ubuntu20.04-py38-cpu"
        environment_version = "1"
        environment_id = _get_environment_arm_id(
            client=client, environment_name=environment_name, environment_version=environment_version
        )
        environment = client._environments.get(name=environment_name, version=environment_version)

        assert environment
        assert environment.id == environment_id

    def test_environment_archive_restore_version(self, client: MLClient, env_name: Callable[[str], str]) -> None:
        versions = ["1", "2"]
        version_archived = versions[0]
        name = env_name("name")
        for version in versions:
            params_override = [{"name": name, "version": version}]
            env = load_environment(
                source="./tests/test_configs/environment/environment_conda.yml", params_override=params_override
            )
            client.environments.create_or_update(env)

        def get_environment_list():
            # Wait for list index to update before calling list command
            sleep_if_live(30)
            environment_list = client.environments.list(name=name, list_view_type=ListViewType.ACTIVE_ONLY)
            return [e.version for e in environment_list if e is not None]

        assert version_archived in get_environment_list()
        client.environments.archive(name=name, version=version_archived)
        assert version_archived not in get_environment_list()
        client.environments.restore(name=name, version=version_archived)
        assert version_archived in get_environment_list()

    def test_environment_archive_restore_container(self, client: MLClient, env_name: Callable[[str], str]) -> None:
        name = env_name("name")
        params_override = [{"name": name}]
        env = load_environment(
            source="./tests/test_configs/environment/environment_conda.yml", params_override=params_override
        )
        client.environments.create_or_update(env)

        def get_environment_list():
            # Wait for list index to update before calling list command
            sleep_if_live(30)
            environment_list = client.environments.list(list_view_type=ListViewType.ACTIVE_ONLY)
            return [e.name for e in environment_list if e is not None]

        assert name in get_environment_list()
        client.environments.archive(name=name)
        assert name not in get_environment_list()
        client.environments.restore(name=name)
        assert name in get_environment_list()

    def test_environment_get_latest_label(self, client: MLClient, randstr: Callable[[], str]) -> None:
        name = randstr("name")
        versions = ["foo", "bar", "baz", "foobar"]

        for version in versions:
            created = client.environments.create_or_update(
                load_environment(
                    "./tests/test_configs/environment/environment_conda_inline.yml",
                    params_override=[{"name": name}, {"version": version}],
                )
            )
            assert created.name == name
            assert created.version == version
            sleep_if_live(2)
            assert client.environments.get(name, label="latest").version == version

    @pytest.mark.live_test_only("Needs re-recording to work with new test proxy sanitizers")
    def test_registry_environment_create_conda_and_get(
        self, only_registry_client: MLClient, env_name: Callable[[str], str]
    ) -> None:
        params_override = [{"name": env_name("name")}]
        env = load_environment(
            source="./tests/test_configs/environment/environment_conda.yml", params_override=params_override
        )

        environment = only_registry_client._environments.create_or_update(env)

        environment_id = f"azureml://registries/testFeed/environments/{env.name}/versions/{env.version}"

        assert environment.conda_file
        assert environment.id == environment_id

        env_dump = environment._to_dict()
        assert env_dump
        assert env_dump["id"] == environment_id
        assert env_dump["conda_file"]
        assert env_dump["description"]

        environment = only_registry_client._environments.get(name=env.name, version=env.version)
        assert environment
        assert environment.id == environment_id

    def test_registry_environment_list(self, only_registry_client: MLClient) -> None:
        environment_list = only_registry_client._environments.list()
        assert environment_list
        assert isinstance(environment_list, ItemPaged)


def _get_environment_arm_id(client: MLClient, environment_name: str, environment_version: str) -> str:
    ws_scope = client._operation_scope
    environment_id = (
        "/subscriptions/{}/resourceGroups/{}/providers/Microsoft.MachineLearningServices/workspaces/{"
        "}/environments/{}/versions/{}".format(
            ws_scope.subscription_id,
            ws_scope.resource_group_name,
            ws_scope.workspace_name,
            environment_name,
            environment_version,
        )
    )

    return environment_id



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from typing import Any

from marshmallow import post_load

from azure.ai.ml._schema._endpoint.batch.batch_endpoint_defaults import BatchEndpointsDefaultsSchema
from azure.ai.ml._schema._endpoint.endpoint import EndpointSchema
from azure.ai.ml._schema.core.fields import NestedField
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY

module_logger = logging.getLogger(__name__)


class BatchEndpointSchema(EndpointSchema):
    defaults = NestedField(BatchEndpointsDefaultsSchema)

    @post_load
    def make(self, data: Any, **kwargs: Any) -> Any:
        from azure.ai.ml.entities import BatchEndpoint

        return BatchEndpoint(base_path=self.context[BASE_PATH_CONTEXT_KEY], **data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging

from marshmallow import fields, post_load

from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY

from .asset import AssetSchema

module_logger = logging.getLogger(__name__)


class WorkspaceAssetReferenceSchema(AssetSchema):
    destination_name = fields.Str()
    destination_version = fields.Str()
    source_asset_id = fields.Str(required=True)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._assets.workspace_asset_reference import WorkspaceAssetReference

        return WorkspaceAssetReference(base_path=self.context[BASE_PATH_CONTEXT_KEY], **data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

"""Contains automated machine learning classes for Azure Machine Learning SDKv2.

Main areas include managing AutoML tasks.
"""
from azure.ai.ml.entities._job.automl import TrainingSettings
from azure.ai.ml.entities._job.automl.image import (
    LogTrainingMetrics,
    LogValidationLoss,
    ImageClassificationJob,
    ImageClassificationMultilabelJob,
    ImageClassificationSearchSpace,
    ImageInstanceSegmentationJob,
    ImageLimitSettings,
    ImageModelSettingsClassification,
    ImageModelSettingsObjectDetection,
    ImageObjectDetectionJob,
    ImageObjectDetectionSearchSpace,
    ImageSweepSettings,
)
from azure.ai.ml.entities._job.automl.nlp import (
    NlpFeaturizationSettings,
    NlpFixedParameters,
    NlpLimitSettings,
    NlpSearchSpace,
    NlpSweepSettings,
    TextClassificationJob,
    TextClassificationMultilabelJob,
    TextNerJob,
)
from azure.ai.ml.entities._job.automl.search_space import SearchSpace
from azure.ai.ml.entities._job.automl.stack_ensemble_settings import (
    StackEnsembleSettings,
)
from azure.ai.ml.entities._job.automl.tabular import (
    ClassificationJob,
    ColumnTransformer,
    ForecastingJob,
    ForecastingSettings,
    RegressionJob,
    TabularFeaturizationSettings,
    TabularLimitSettings,
)

from .._restclient.v2023_04_01_preview.models import (
    BlockedTransformers,
    ClassificationModels,
    ClassificationMultilabelPrimaryMetrics,
    ClassificationPrimaryMetrics,
    FeaturizationMode,
    ForecastHorizonMode,
    ForecastingModels,
    ForecastingPrimaryMetrics,
    InstanceSegmentationPrimaryMetrics,
    LearningRateScheduler,
    NCrossValidationsMode,
    ObjectDetectionPrimaryMetrics,
    RegressionModels,
    RegressionPrimaryMetrics,
    SamplingAlgorithmType,
    ShortSeriesHandlingConfiguration,
    StochasticOptimizer,
    TargetAggregationFunction,
    TargetLagsMode,
    TargetRollingWindowSizeMode,
    UseStl,
    ValidationMetricType,
)
from ._automl_image import (
    image_classification,
    image_classification_multilabel,
    image_instance_segmentation,
    image_object_detection,
)
from ._automl_nlp import text_classification, text_classification_multilabel, text_ner
from ._automl_tabular import classification, forecasting, regression

__all__ = [
    "ClassificationModels",
    "RegressionModels",
    "ForecastingModels",
    "FeaturizationMode",
    "NCrossValidationsMode",
    "ForecastHorizonMode",
    "ShortSeriesHandlingConfiguration",
    "TargetLagsMode",
    "TargetRollingWindowSizeMode",
    "TargetAggregationFunction",
    "UseStl",
    "ClassificationPrimaryMetrics",
    "RegressionPrimaryMetrics",
    "ForecastingPrimaryMetrics",
    "ClassificationMultilabelPrimaryMetrics",
    "ObjectDetectionPrimaryMetrics",
    "InstanceSegmentationPrimaryMetrics",
    "ColumnTransformer",
    "TabularFeaturizationSettings",
    "ForecastingSettings",
    "TabularLimitSettings",
    "NlpFeaturizationSettings",
    "NlpFixedParameters",
    "NlpLimitSettings",
    "NlpSweepSettings",
    "NlpSearchSpace",
    "LogTrainingMetrics",
    "LogValidationLoss",
    "ImageLimitSettings",
    "ImageModelSettingsClassification",
    "ImageModelSettingsObjectDetection",
    "ImageSweepSettings",
    "ImageObjectDetectionSearchSpace",
    "ImageClassificationSearchSpace",
    "TrainingSettings",
    "image_classification",
    "image_classification_multilabel",
    "image_object_detection",
    "image_instance_segmentation",
    "text_classification",
    "text_classification_multilabel",
    "text_ner",
    "classification",
    "regression",
    "forecasting",
    "SearchSpace",
    "StackEnsembleSettings",
    "BlockedTransformers",
    "ClassificationJob",
    "ForecastingJob",
    "RegressionJob",
    "ImageClassificationJob",
    "ImageClassificationMultilabelJob",
    "ImageObjectDetectionJob",
    "ImageInstanceSegmentationJob",
    "LearningRateScheduler",
    "SamplingAlgorithmType",
    "StochasticOptimizer",
    "TextClassificationJob",
    "TextClassificationMultilabelJob",
    "TextNerJob",
    "ValidationMetricType",
]



import os
from pathlib import Path

import pytest
import yaml
from marshmallow.exceptions import ValidationError

from azure.ai.ml import load_data
from azure.ai.ml.entities._assets import Data
from azure.ai.ml.entities._assets.asset import Asset
from azure.ai.ml.entities._assets.auto_delete_setting import AutoDeleteSetting, AutoDeleteCondition


@pytest.mark.unittest
@pytest.mark.data_experiences_test
class TestData:
    def test_deserialize_file(self):
        test_path = "./tests/test_configs/dataset/data_file.yaml"
        with open(test_path, "r") as f:
            target = yaml.safe_load(f)
        with open(test_path, "r") as f:
            data: Data = load_data(source=test_path)
        assert data.name == target["name"]
        source = data._to_rest_object()
        assert os.path.normpath(source.properties.data_uri) == os.path.normpath(
            Path(Path(test_path).parent, target["path"]).resolve()
        )
        assert source.properties.data_type == "uri_file"

    def test_deserialize_folder(self):
        test_path = "./tests/test_configs/dataset/data_local_path.yaml"
        with open(test_path, "r") as f:
            target = yaml.safe_load(f)
        with open(test_path, "r") as f:
            data: Data = load_data(source=test_path)
        assert data.name == target["name"]
        source = data._to_rest_object()
        assert os.path.normpath(source.properties.data_uri) == os.path.normpath(
            Path(Path(test_path).parent, target["path"]).resolve()
        )
        assert source.properties.data_type == "uri_folder"

    def test_deserialize_mltable(self):
        test_path = "./tests/test_configs/dataset/data_local_path_mltable.yaml"
        with open(test_path, "r") as f:
            target = yaml.safe_load(f)
        with open(test_path, "r") as f:
            dataset: Data = load_data(source=test_path)
        assert dataset.name == target["name"]
        source = dataset._to_rest_object()
        assert os.path.normpath(source.properties.data_uri) == os.path.normpath(
            Path(Path(test_path).parent, target["path"]).resolve()
        )
        assert source.properties.data_type == "mltable"

    def test_deserialize_empty_paths_should_get_validation_error(self):
        test_path = "./tests/test_configs/dataset/data_missing_path_test.yml"
        with open(test_path, "r") as f:
            yaml.safe_load(f)
        with open(test_path, "r"):
            with pytest.raises(ValidationError) as e:
                load_data(source=test_path)
        assert "Missing data for required field" in e.value.messages[0]

    def test_asset_eq(self):
        auto_delete_setting = AutoDeleteSetting(condition=AutoDeleteCondition.CREATED_GREATER_THAN, value="30d")
        # with auto delete setting
        targetAsset = Data(
            name="testDataAssetfolder",
            version="1",
            description="this is a test data asset with auto delete setting",
            path="azureml://datastores/workspacemanageddatastore/",
            auto_delete_setting=auto_delete_setting,
        )
        test_path = "./tests/test_configs/dataset/data_with_auto_delete_setting.yml"
        with open(test_path, "r") as f:
            asset: Data = load_data(source=test_path)

            # the path types are different in load_data and raw Data
            targetAsset._base_path = asset.base_path
            assert Asset.__eq__(asset, targetAsset) is True

        # without auto delete setting
        targetAsset = Data(
            name="testDataAssetfolder",
            version="1",
            description="this is a test data asset without auto delete setting",
            path="azureml://datastores/workspacemanageddatastore/",
            auto_delete_setting=auto_delete_setting,
        )
        test_path = "./tests/test_configs/dataset/data_without_auto_delete_setting.yml"
        with open(test_path, "r") as f:
            asset: Data = load_data(source=test_path)

            # the path types are different in load_data and raw Data
            targetAsset._base_path = asset.base_path
            assert Asset.__eq__(asset, targetAsset) is False



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import yaml

from azure.ai.ml._schema.core.fields import NestedField, FileRefField
from azure.ai.ml._schema.schedule.schedule import ScheduleSchema
from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY
from ..core.fields import UnionField
from .data_import import DataImportSchema


class ImportDataFileRefField(FileRefField):
    def _deserialize(self, value, attr, data, **kwargs) -> "DataImport":
        # Get component info from component yaml file.
        data = super()._deserialize(value, attr, data, **kwargs)
        data_import_dict = yaml.safe_load(data)

        from azure.ai.ml.entities._data_import.data_import import DataImport

        return DataImport._load(
            data=data_import_dict,
            yaml_path=self.context[BASE_PATH_CONTEXT_KEY] / value,
            **kwargs,
        )


@experimental
class ImportDataScheduleSchema(ScheduleSchema):
    import_data = UnionField(
        [
            ImportDataFileRefField,
            NestedField(DataImportSchema),
        ]
    )



# imports
import argparse

import lightgbm as lgbm
import matplotlib.pyplot as plt
import mlflow
import pandas as pd
from sklearn.metrics import accuracy_score, log_loss
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder


# define functions
def main(args):
    # enable auto logging
    mlflow.autolog()

    # setup parameters
    num_boost_round = args.num_boost_round
    params = {
        "objective": "multiclass",
        "num_class": 3,
        "boosting": args.boosting,
        "num_iterations": args.num_iterations,
        "num_leaves": args.num_leaves,
        "num_threads": args.num_threads,
        "learning_rate": args.learning_rate,
        "metric": args.metric,
        "seed": args.seed,
        "verbose": args.verbose,
    }

    # read in data
    df = pd.read_csv(args.iris_csv)

    # process data
    X_train, X_test, y_train, y_test, enc = process_data(df)

    # train model
    _ = train_model(params, num_boost_round, X_train, X_test, y_train, y_test)


def process_data(df):
    # split dataframe into X and y
    X = df.drop(["species"], axis=1)
    y = df["species"]

    # encode label
    enc = LabelEncoder()
    y = enc.fit_transform(y)

    # train/test split
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

    # return splits and encoder
    return X_train, X_test, y_train, y_test, enc


def train_model(params, num_boost_round, X_train, X_test, y_train, y_test):
    # create lightgbm datasets
    train_data = lgbm.Dataset(X_train, label=y_train)
    test_data = lgbm.Dataset(X_test, label=y_test)

    # train model
    model = lgbm.train(
        params,
        train_data,
        num_boost_round=num_boost_round,
        valid_sets=[test_data],
        valid_names=["test"],
    )

    # return model
    return model


def parse_args():
    # setup arg parser
    parser = argparse.ArgumentParser()

    # add arguments
    parser.add_argument("--iris-csv", type=str)
    parser.add_argument("--num-boost-round", type=int, default=10)
    parser.add_argument("--boosting", type=str, default="gbdt")
    parser.add_argument("--num-iterations", type=int, default=16)
    parser.add_argument("--num-leaves", type=int, default=31)
    parser.add_argument("--num-threads", type=int, default=0)
    parser.add_argument("--learning-rate", type=float, default=0.1)
    parser.add_argument("--metric", type=str, default="multi_logloss")
    parser.add_argument("--seed", type=int, default=42)
    parser.add_argument("--verbose", type=int, default=0)

    # parse args
    args = parser.parse_args()

    # return args
    return args


# run script
if __name__ == "__main__":
    # parse args
    args = parse_args()

    # run main function
    main(args)



from pathlib import Path

from azure.ai.ml import Input, Output, dsl, load_component
from azure.ai.ml.data_transfer import copy_data
from azure.ai.ml.constants._common import AssetTypes, SERVERLESS_COMPUTE
from azure.ai.ml.constants._component import DataTransferTaskType, DataCopyMode
from azure.ai.ml.entities import PipelineJob


parent_dir = str(Path(__file__).parent)


def generate_dsl_pipeline_from_yaml() -> PipelineJob:
    merge_files_func = load_component(parent_dir + "/merge_files.yaml")

    @dsl.pipeline(description="submit a pipeline with data transfer copy job")
    def data_transfer_copy_pipeline_from_yaml(cosmos_folder, cosmos_folder_dup):
        merge_files = merge_files_func(folder1=cosmos_folder, folder2=cosmos_folder_dup)
        return {"merged_blob": merge_files.outputs.output_folder}

    pipeline = data_transfer_copy_pipeline_from_yaml(
        cosmos_folder=Input(
            path=parent_dir + "/data/iris_model",
            type=AssetTypes.URI_FOLDER,
        ),
        cosmos_folder_dup=Input(
            type=AssetTypes.URI_FOLDER,
            path=parent_dir + "/data/iris_model",
        ),
    )

    pipeline.settings.default_compute = SERVERLESS_COMPUTE
    return pipeline


def generate_dsl_pipeline_from_builder() -> PipelineJob:
    cosmos_folder = Input(
        path=parent_dir + "/data/iris_model",
        type=AssetTypes.URI_FOLDER,
    )
    cosmos_folder_dup = Input(
        path=parent_dir + "/data/iris_model",
        type=AssetTypes.URI_FOLDER,
    )

    inputs = {
        "folder1": cosmos_folder,
        "folder2": cosmos_folder_dup,
    }
    outputs = {"output_folder": Output(type=AssetTypes.URI_FOLDER)}

    merge_files_func = copy_data(
        inputs=inputs,
        outputs=outputs,
        data_copy_mode=DataCopyMode.MERGE_WITH_OVERWRITE,
    )

    # Define pipeline
    @dsl.pipeline(description="submit a pipeline with data transfer copy job")
    def data_transfer_copy_pipeline_from_builder(cosmos_folder, cosmos_folder_dup):
        merge_files = merge_files_func(folder1=cosmos_folder, folder2=cosmos_folder_dup)
        return {"merged_blob": merge_files.outputs.output_folder}

    pipeline = data_transfer_copy_pipeline_from_builder(
        cosmos_folder=cosmos_folder, cosmos_folder_dup=cosmos_folder_dup
    )
    pipeline.settings.default_compute = SERVERLESS_COMPUTE
    return pipeline


def generate_dsl_pipeline_copy_mixtype_from_yaml() -> PipelineJob:
    merge_files_func = load_component(parent_dir + "/merge_mixtype_files.yaml")

    @dsl.pipeline(description="submit a pipeline with data transfer copy job")
    def data_transfer_copy_pipeline_from_yaml(input1, input2, input3):
        merge_files = merge_files_func(input1=input1, input2=input2, input3=input3)
        return {"merged_blob": merge_files.outputs.output_folder}

    pipeline = data_transfer_copy_pipeline_from_yaml(
        input1=Input(
            path=parent_dir + "/data/sample.csv",
            type=AssetTypes.URI_FILE,
        ),
        input2=Input(
            path=parent_dir + "/data/sample.csv",
            type=AssetTypes.URI_FILE,
        ),
        input3=Input(
            path=parent_dir + "/data/iris-mltable",
            type=AssetTypes.MLTABLE,
        ),
    )

    pipeline.settings.default_compute = SERVERLESS_COMPUTE
    return pipeline


def generate_dsl_pipeline_copy_urifile_from_yaml() -> PipelineJob:
    merge_files_func = load_component(parent_dir + "/copy_uri_files.yaml")

    @dsl.pipeline(description="submit a pipeline with data transfer copy job")
    def data_transfer_copy_pipeline_from_yaml(input1):
        merge_files = merge_files_func(input1=input1)
        return {"merged_blob": merge_files.outputs.output_folder}

    pipeline = data_transfer_copy_pipeline_from_yaml(
        input1=Input(
            path=parent_dir + "/data/sample.csv",
            type=AssetTypes.URI_FILE,
        ),
    )

    # pipeline.outputs.merged_blob.path will convert type to uri_folder, has a bug to track this: 2104275
    pipeline.settings.default_compute = SERVERLESS_COMPUTE
    return pipeline


def generate_dsl_pipeline_copy_urifolder_from_yaml() -> PipelineJob:
    merge_files_func = load_component(parent_dir + "/copy_files.yaml")

    @dsl.pipeline(description="submit a pipeline with data transfer copy job")
    def data_transfer_copy_pipeline_from_yaml(input1):
        merge_files = merge_files_func(input1=input1)
        return {"merged_blob": merge_files.outputs.output_folder}

    pipeline = data_transfer_copy_pipeline_from_yaml(
        input1=Input(
            path=parent_dir + "/data/iris_model",
            type=AssetTypes.URI_FOLDER,
        ),
    )
    pipeline.settings.default_compute = SERVERLESS_COMPUTE
    return pipeline



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from marshmallow import EXCLUDE, fields

from azure.ai.ml._schema._utils.utils import validate_arm_str
from azure.ai.ml._schema.core.fields import NestedField, StringTransformedEnum
from azure.ai.ml._schema.core.schema import PathAwareSchema
from azure.ai.ml._schema.workspace.customer_managed_key import CustomerManagedKeySchema
from azure.ai.ml._schema.workspace.identity import IdentitySchema
from azure.ai.ml._schema.workspace.network_acls import NetworkAclsSchema
from azure.ai.ml._schema.workspace.networking import ManagedNetworkSchema
from azure.ai.ml._schema.workspace.serverless_compute import ServerlessComputeSettingsSchema
from azure.ai.ml._utils.utils import snake_to_pascal
from azure.ai.ml.constants._common import PublicNetworkAccess


class WorkspaceSchema(PathAwareSchema):
    name = fields.Str(required=True)
    location = fields.Str()
    id = fields.Str(dump_only=True)
    resource_group = fields.Str()
    description = fields.Str()
    discovery_url = fields.Str()
    display_name = fields.Str()
    hbi_workspace = fields.Bool()
    storage_account = fields.Str(validate=validate_arm_str)
    container_registry = fields.Str(validate=validate_arm_str)
    key_vault = fields.Str(validate=validate_arm_str)
    application_insights = fields.Str(validate=validate_arm_str)
    customer_managed_key = NestedField(CustomerManagedKeySchema)
    tags = fields.Dict(keys=fields.Str(), values=fields.Str())
    mlflow_tracking_uri = fields.Str(dump_only=True)
    image_build_compute = fields.Str()
    public_network_access = StringTransformedEnum(
        allowed_values=[PublicNetworkAccess.DISABLED, PublicNetworkAccess.ENABLED],
        casing_transform=snake_to_pascal,
    )
    network_acls = NestedField(NetworkAclsSchema)
    system_datastores_auth_mode = fields.Str()
    identity = NestedField(IdentitySchema)
    primary_user_assigned_identity = fields.Str()
    workspace_hub = fields.Str(validate=validate_arm_str)
    managed_network = NestedField(ManagedNetworkSchema, unknown=EXCLUDE)
    provision_network_now = fields.Bool()
    enable_data_isolation = fields.Bool()
    allow_roleassignment_on_rg = fields.Bool()
    serverless_compute = NestedField(ServerlessComputeSettingsSchema)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
from typing import Any

from marshmallow import post_load, fields

from azure.ai.ml._schema import PatchedSchemaMeta, StringTransformedEnum, NestedField, UnionField
from azure.ai.ml._schema._deployment.online.data_asset_schema import DataAssetSchema
from azure.ai.ml.constants._common import Boolean

module_logger = logging.getLogger(__name__)


class DeploymentCollectionSchema(metaclass=PatchedSchemaMeta):
    enabled = StringTransformedEnum(required=True, allowed_values=[Boolean.TRUE, Boolean.FALSE])
    data = UnionField(
        [
            NestedField(DataAssetSchema),
            fields.Str(),
        ]
    )
    client_id = fields.Str()

    # pylint: disable=unused-argument
    @post_load
    def make(self, data: Any, **kwargs: Any) -> Any:
        from azure.ai.ml.entities._deployment.deployment_collection import DeploymentCollection

        return DeploymentCollection(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint: disable=protected-access
from typing import Dict, List, Optional, Union

from marshmallow import INCLUDE, Schema

from ... import MpiDistribution, PyTorchDistribution, RayDistribution, TensorFlowDistribution
from ..._schema import PathAwareSchema
from ..._schema.core.fields import DistributionField
from ...entities import CommandJobLimits, JobResourceConfiguration
from ...entities._util import get_rest_dict_for_node_attrs
from .._schema.component import NodeType
from ..entities.component import InternalComponent
from ..entities.node import InternalBaseNode


class Command(InternalBaseNode):
    """Node of internal command components in pipeline with specific run settings.

    Different from azure.ai.ml.entities.Command, type of this class is CommandComponent.
    """

    def __init__(self, **kwargs):
        node_type = kwargs.pop("type", None) or NodeType.COMMAND
        super(Command, self).__init__(type=node_type, **kwargs)
        self._init = True
        self._resources = kwargs.pop("resources", JobResourceConfiguration())
        self._compute = kwargs.pop("compute", None)
        self._environment = kwargs.pop("environment", None)
        self._environment_variables = kwargs.pop("environment_variables", None)
        self._limits = kwargs.pop("limits", CommandJobLimits())
        self._init = False

    @property
    def compute(self) -> Optional[str]:
        """Get the compute definition for the command.

        :return: The compute definition
        :rtype: Optional[str]
        """
        return self._compute

    @compute.setter
    def compute(self, value: str) -> None:
        """Set the compute definition for the command.

        :param value: The new compute definition
        :type value: str
        """
        self._compute = value

    @property
    def environment(self) -> Optional[str]:
        """Get the environment definition for the command.

        :return: The environment definition
        :rtype: Optional[str]
        """
        return self._environment

    @environment.setter
    def environment(self, value: str) -> None:
        """Set the environment definition for the command.

        :param value: The new environment definition
        :type value: str
        """
        self._environment = value

    @property
    def environment_variables(self) -> Optional[Dict[str, str]]:
        """Get the environment variables for the command.

        :return: The environment variables
        :rtype: Optional[Dict[str, str]]
        """
        return self._environment_variables

    @environment_variables.setter
    def environment_variables(self, value: Dict[str, str]) -> None:
        """Set the environment variables for the command.

        :param value: The new environment variables
        :type value: Dict[str, str]
        """
        self._environment_variables = value

    @property
    def limits(self) -> CommandJobLimits:
        return self._limits

    @limits.setter
    def limits(self, value: CommandJobLimits):
        self._limits = value

    @property
    def resources(self) -> JobResourceConfiguration:
        """Compute Resource configuration for the component.

        :return: The resource configuration
        :rtype: JobResourceConfiguration
        """
        return self._resources

    @resources.setter
    def resources(self, value: JobResourceConfiguration):
        self._resources = value

    @classmethod
    def _picked_fields_from_dict_to_rest_object(cls) -> List[str]:
        return ["environment", "limits", "resources", "environment_variables"]

    @classmethod
    def _create_schema_for_validation(cls, context) -> Union[PathAwareSchema, Schema]:
        from .._schema.command import CommandSchema

        return CommandSchema(context=context)

    def _to_rest_object(self, **kwargs) -> dict:
        rest_obj = super()._to_rest_object(**kwargs)
        rest_obj.update(
            {
                "limits": get_rest_dict_for_node_attrs(self.limits, clear_empty_value=True),
                "resources": get_rest_dict_for_node_attrs(self.resources, clear_empty_value=True),
            }
        )
        return rest_obj

    @classmethod
    def _from_rest_object_to_init_params(cls, obj):
        obj = InternalBaseNode._from_rest_object_to_init_params(obj)

        if "resources" in obj and obj["resources"]:
            obj["resources"] = JobResourceConfiguration._from_rest_object(obj["resources"])

        # handle limits
        if "limits" in obj and obj["limits"]:
            obj["limits"] = CommandJobLimits._from_rest_object(obj["limits"])
        return obj


class Distributed(Command):
    def __init__(self, **kwargs):
        super(Distributed, self).__init__(**kwargs)
        self._distribution = kwargs.pop("distribution", None)
        self._type = NodeType.DISTRIBUTED
        if self._distribution is None:
            # hack: distribution.type is required to set distribution, which is defined in launcher.type
            if (
                isinstance(self.component, InternalComponent)
                and self.component.launcher
                and "type" in self.component.launcher
            ):
                self.distribution = {"type": self.component.launcher["type"]}
            else:
                raise ValueError(
                    "launcher.type must be specified in definition of DistributedComponent but got {}".format(
                        self.component
                    )
                )

    @property
    def distribution(
        self,
    ) -> Union[PyTorchDistribution, MpiDistribution, TensorFlowDistribution, RayDistribution]:
        """The distribution config of component, e.g. distribution={'type': 'mpi'}.

        :return: The distribution config
        :rtype: Union[PyTorchDistribution, MpiDistribution, TensorFlowDistribution, RayDistribution]
        """
        return self._distribution

    @distribution.setter
    def distribution(
        self,
        value: Union[Dict, PyTorchDistribution, TensorFlowDistribution, MpiDistribution, RayDistribution],
    ):
        if isinstance(value, dict):
            dist_schema = DistributionField(unknown=INCLUDE)
            value = dist_schema._deserialize(value=value, attr=None, data=None)
        self._distribution = value

    @classmethod
    def _create_schema_for_validation(cls, context) -> Union[PathAwareSchema, Schema]:
        from .._schema.command import DistributedSchema

        return DistributedSchema(context=context)

    @classmethod
    def _picked_fields_from_dict_to_rest_object(cls) -> List[str]:
        return Command._picked_fields_from_dict_to_rest_object() + ["distribution"]

    def _to_rest_object(self, **kwargs) -> dict:
        rest_obj = super()._to_rest_object(**kwargs)
        distribution = self.distribution._to_rest_object() if self.distribution else None  # pylint: disable=no-member
        rest_obj.update(
            {
                "distribution": get_rest_dict_for_node_attrs(distribution),
            }
        )
        return rest_obj



from typing import Any, List, Tuple

import pytest

from azure.ai.ml._restclient.v2023_04_01_preview.models import (
    AutoForecastHorizon,
    AutoSeasonality,
    AutoTargetLags,
    AutoTargetRollingWindowSize,
    CustomForecastHorizon,
    CustomSeasonality,
    CustomTargetLags,
    CustomTargetRollingWindowSize,
)
from azure.ai.ml._restclient.v2023_04_01_preview.models import ForecastingSettings as RestForecastingSettings
from azure.ai.ml._restclient.v2023_04_01_preview.models import ShortSeriesHandlingConfiguration
from azure.ai.ml.entities._job.automl.tabular.forecasting_settings import ForecastingSettings


@pytest.mark.automl_test
@pytest.mark.unittest
class TestForecastingSettings:
    @pytest.mark.parametrize(
        "settings",
        [
            RestForecastingSettings(time_series_id_column_names=["xxx"]),
            RestForecastingSettings(time_series_id_column_names=[]),
            RestForecastingSettings(time_series_id_column_names=None),
        ],
    )
    def test_forecast_settings_from_object(self, settings: RestForecastingSettings):
        actual_obj = ForecastingSettings._from_rest_object(obj=settings)
        assert actual_obj.time_series_id_column_names == settings.time_series_id_column_names

    @pytest.mark.parametrize(
        "settings",
        [
            (ForecastingSettings(time_series_id_column_names=["xxx"]), ["xxx"]),
            (ForecastingSettings(time_series_id_column_names=[]), []),
            (ForecastingSettings(time_series_id_column_names=None), None),
            (ForecastingSettings(time_series_id_column_names="xxx"), ["xxx"]),
        ],
    )
    def test_forecast_settings_to_rest_object(self, settings: Tuple[ForecastingSettings, List[str]]):
        input_settings, expected = settings
        actual = input_settings._to_rest_object()
        assert actual.time_series_id_column_names == expected

    @pytest.mark.parametrize(
        "settings",
        [
            (RestForecastingSettings(forecast_horizon=CustomForecastHorizon(value=10)), 10),
            (RestForecastingSettings(forecast_horizon=AutoForecastHorizon()), "auto"),
            (RestForecastingSettings(forecast_horizon=None), None),
        ],
    )
    def test_forecast_horizon_from_rest(self, settings: Tuple[RestForecastingSettings, Any]):
        actual_obj = ForecastingSettings._from_rest_object(obj=settings[0])
        assert actual_obj.forecast_horizon == settings[1], "actual: {}, expected: {}".format(
            actual_obj.forecast_horizon, settings[1]
        )

    @pytest.mark.parametrize(
        "settings",
        [
            (ForecastingSettings(forecast_horizon=10), CustomForecastHorizon(value=10)),
            (ForecastingSettings(forecast_horizon="auto"), AutoForecastHorizon()),
            (ForecastingSettings(forecast_horizon=None), None),
        ],
    )
    def test_forecast_horizon_to_rest(self, settings: Tuple[ForecastingSettings, List[str]]):
        input_settings, expected = settings
        actual = input_settings._to_rest_object()
        assert actual.forecast_horizon == expected

    # test target_lags from rest
    @pytest.mark.parametrize(
        "settings",
        [
            (RestForecastingSettings(target_lags=CustomTargetLags(values=[10, 20])), [10, 20]),
            (RestForecastingSettings(target_lags=AutoTargetLags()), "auto"),
            (RestForecastingSettings(target_lags=None), None),
        ],
    )
    def test_target_lags_from_rest(self, settings: Tuple[RestForecastingSettings, Any]):
        actual_obj = ForecastingSettings._from_rest_object(obj=settings[0])
        assert actual_obj.target_lags == settings[1], "actual: {}, expected: {}".format(
            actual_obj.target_lags, settings[1]
        )

    @pytest.mark.parametrize(
        "settings",
        [
            (ForecastingSettings(target_lags=10), CustomTargetLags(values=[10])),
            (ForecastingSettings(target_lags=[10, 20, 30]), CustomTargetLags(values=[10, 20, 30])),
            (ForecastingSettings(target_lags="auto"), AutoTargetLags()),
        ],
    )
    def test_target_lags_to_rest(self, settings: Tuple[ForecastingSettings, List[int]]):
        input_settings, expected = settings
        actual = input_settings._to_rest_object()
        assert actual.target_lags == expected, "actual: {}, expected: {}".format(actual, expected)

    # test target_rolling_window_size from rest
    @pytest.mark.parametrize(
        "settings",
        [
            (RestForecastingSettings(target_rolling_window_size=CustomTargetRollingWindowSize(value=10)), 10),
            (RestForecastingSettings(target_rolling_window_size=AutoTargetRollingWindowSize()), "auto"),
            (RestForecastingSettings(target_rolling_window_size=None), None),
        ],
    )
    def test_target_rolling_window_size_from_rest(self, settings: Tuple[RestForecastingSettings, Any]):
        actual_obj = ForecastingSettings._from_rest_object(obj=settings[0])
        assert actual_obj.target_rolling_window_size == settings[1], "actual: {}, expected: {}".format(
            actual_obj.target_rolling_window_size, settings[1]
        )

    @pytest.mark.parametrize(
        "settings",
        [
            (ForecastingSettings(target_rolling_window_size=10), CustomTargetRollingWindowSize(value=10)),
            (ForecastingSettings(target_rolling_window_size="auto"), AutoTargetRollingWindowSize()),
        ],
    )
    def test_target_rolling_window_size_to_rest(self, settings: Tuple[ForecastingSettings, List[int]]):
        input_settings, expected = settings
        actual = input_settings._to_rest_object()
        assert actual.target_rolling_window_size == expected, "actual: {}, expected: {}".format(actual, expected)

    # test seasonality from rest
    @pytest.mark.parametrize(
        "settings",
        [
            (RestForecastingSettings(seasonality=CustomSeasonality(value=10)), 10),
            (RestForecastingSettings(seasonality=AutoSeasonality()), "auto"),
            (RestForecastingSettings(seasonality=None), None),
        ],
    )
    def test_seasonality_from_rest(self, settings: Tuple[RestForecastingSettings, Any]):
        actual_obj = ForecastingSettings._from_rest_object(obj=settings[0])
        assert actual_obj.seasonality == settings[1], "actual: {}, expected: {}".format(
            actual_obj.seasonality, settings[1]
        )

    @pytest.mark.parametrize(
        "settings",
        [
            (ForecastingSettings(seasonality=10), CustomSeasonality(value=10)),
            (ForecastingSettings(seasonality="auto"), AutoSeasonality()),
        ],
    )
    def test_seasonality_to_rest(self, settings: Tuple[ForecastingSettings, List[int]]):
        input_settings, expected = settings
        actual = input_settings._to_rest_object()
        assert actual.seasonality == expected, "actual: {}, expected: {}".format(actual, expected)

    def test_forecast_settings_equality(self):
        settings = ForecastingSettings(
            time_column_name="time",
            forecast_horizon=20,
            target_lags=[1, 2, 3],
            time_series_id_column_names=["tid1", "tid2", "tid2"],
            frequency="W-THU",
            target_rolling_window_size=4,
            short_series_handling_config=ShortSeriesHandlingConfiguration.DROP,
            use_stl="season",
            seasonality=3,
            features_unknown_at_forecast_time=["a"],
        )

        # serialize and deserialize again
        settings_rest = settings._to_rest_object()
        settings_rest_2 = ForecastingSettings._from_rest_object(obj=settings_rest)
        assert settings == settings_rest_2, "actual: {}, expected: {}".format(settings, settings_rest_2)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------


import logging
from typing import Any

from marshmallow import fields, post_load, validates, ValidationError

from azure.ai.ml._schema import NestedField, PatchedSchemaMeta, StringTransformedEnum
from azure.ai.ml._schema._deployment.online.request_logging_schema import RequestLoggingSchema
from azure.ai.ml._schema._deployment.online.deployment_collection_schema import DeploymentCollectionSchema

from azure.ai.ml.constants._common import RollingRate

module_logger = logging.getLogger(__name__)


class DataCollectorSchema(metaclass=PatchedSchemaMeta):
    collections = fields.Dict(keys=fields.Str, values=NestedField(DeploymentCollectionSchema))
    rolling_rate = StringTransformedEnum(
        required=False,
        allowed_values=[RollingRate.MINUTE, RollingRate.DAY, RollingRate.HOUR],
    )
    sampling_rate = fields.Float()  # Should be copied to each of the collections
    request_logging = NestedField(RequestLoggingSchema)

    # pylint: disable=unused-argument
    @validates("sampling_rate")
    def validate_sampling_rate(self, value, **kwargs):
        if value > 1.0 or value < 0.0:
            raise ValidationError("Sampling rate must be an number in range (0.0-1.0)")

    @post_load
    def make(self, data: Any, **kwargs: Any) -> Any:  # pylint: disable=unused-argument
        from azure.ai.ml.entities._deployment.data_collector import DataCollector

        return DataCollector(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
import pytest
from devtools_testutils import AzureRecordedTestCase

from .._util import _UTILS_TIMEOUT_SECOND


@pytest.mark.timeout(_UTILS_TIMEOUT_SECOND)
@pytest.mark.e2etest
@pytest.mark.usefixtures("recorded_test")
@pytest.mark.core_sdk_test
class TestTelemetryValue(AzureRecordedTestCase):
    def test_component_node_telemetry_value(self, hello_world_component):
        # From remote
        v = hello_world_component._get_telemetry_values()
        assert v["type"] == "command"
        assert v["source"] == "REMOTE.WORKSPACE.COMPONENT"
        assert v["is_anonymous"] is False
        component = hello_world_component()
        v = component._get_telemetry_values()
        assert v["type"] == "command"
        assert v["source"] == "REMOTE.WORKSPACE.COMPONENT"
